import tempfile
from pathlib import Path
from unittest.mock import patch

import pytest
from typer.testing import CliRunner

from oumi.cli.cli_utils import CONTEXT_ALLOW_EXTRA_ARGS
from oumi.cli.judge import judge_dataset_file
from oumi.judges.base_judge import JudgeOutput

runner = CliRunner()


@pytest.fixture
def mock_parse_extra_cli_args():
    with patch("oumi.cli.cli_utils.parse_extra_cli_args") as m_parse:
        m_parse.return_value = {}
        yield m_parse


@pytest.fixture
def app():
    import typer

    judge_app = typer.Typer()
    judge_app.command(context_settings=CONTEXT_ALLOW_EXTRA_ARGS)(judge_dataset_file)
    yield judge_app


@pytest.fixture
def mock_judge_file():
    with patch("oumi.judge.judge_dataset_file") as m_jf:
        yield m_jf


@pytest.fixture
def mock_judge_config_from_path():
    with patch("oumi.core.configs.judge_config.JudgeConfig.from_path") as m_rjc:
        yield m_rjc


@pytest.fixture
def sample_judge_output():
    return JudgeOutput(
        raw_output="Test judgment",
        parsed_output={"quality": "good"},
        field_values={"quality": "good"},
        field_scores={"quality": 0.5},
    )


def test_judge_file(
    app,
    mock_parse_extra_cli_args,
    mock_judge_file,
    mock_judge_config_from_path,
    sample_judge_output,
):
    """Test that judge_file command runs successfully with all required parameters."""
    judge_config = "judge_config.yaml"
    input_file = "input.jsonl"

    mock_judge_file.return_value = [sample_judge_output]

    with patch("oumi.cli.judge.Path") as mock_path:
        mock_path.return_value.exists.return_value = True
        result = runner.invoke(
            app,
            [
                "dataset",
                "--config",
                judge_config,
                "--input",
                input_file,
            ],
        )

        assert result.exit_code == 0
        mock_parse_extra_cli_args.assert_called_once()
        mock_judge_config_from_path.assert_called_once_with(
            path=judge_config, extra_args={}
        )

        mock_judge_file.assert_called_once_with(
            judge_config=mock_judge_config_from_path.return_value,
            input_file=input_file,
            output_file=None,
        )


def test_judge_file_with_output_file(
    app,
    mock_parse_extra_cli_args,
    mock_judge_file,
    mock_judge_config_from_path,
    sample_judge_output,
):
    """Test that judge_file saves results to output file when specified."""
    with tempfile.TemporaryDirectory() as temp_dir:
        judge_config = "judge_config.yaml"
        input_file = "input.jsonl"
        output_file = str(Path(temp_dir) / "output.jsonl")

        mock_judge_file.return_value = [sample_judge_output]

        with patch("oumi.cli.judge.Path") as mock_path:
            mock_path.return_value.exists.return_value = True
            result = runner.invoke(
                app,
                [
                    "dataset",
                    "--config",
                    judge_config,
                    "--input",
                    input_file,
                    "--output",
                    output_file,
                ],
            )

            assert result.exit_code == 0
            mock_parse_extra_cli_args.assert_called_once()
            mock_judge_config_from_path.assert_called_once_with(
                path=judge_config, extra_args={}
            )

            mock_judge_file.assert_called_once_with(
                judge_config=mock_judge_config_from_path.return_value,
                input_file=input_file,
                output_file=output_file,
            )



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
from typing import Any, NamedTuple, Optional

from oumi.core.tokenizers.base_tokenizer import BaseTokenizer
from oumi.utils.debug_utils import log_example_for_debugging
from oumi.utils.logging import logger
from oumi.utils.torch_utils import (
    create_ones_like,
    pad_sequences,
    pad_to_max_dim_and_stack,
)

_INPUT_IDS_KEY = "input_ids"
_ATTENTION_MASK_KEY = "attention_mask"
_CROSS_ATTENTION_MASK_KEY = "cross_attention_mask"
_LABELS_KEY = "labels"


class _SpecialTokens(NamedTuple):
    """Special tokens used by VisionLanguageCollatorWithPadding."""

    pad_token_id: int
    """Token id of `PAD` token."""

    label_ignore_index: Optional[int]
    """If set, then `PAD` tokens will be replaced by this special value
    to exclude them from the loss computation.
    """


class TextCollatorWithPadding:
    def __init__(
        self,
        tokenizer: BaseTokenizer,
        *,
        max_length: Optional[int],
        truncation: bool = False,
        label_ignore_index: Optional[int] = None,
        max_variable_sized_dims: int = 1,
        debug: bool = False,
    ):
        """Custom collator for text LLM training.

        Args:
        tokenizer: The tokenizer used for encoding the data.
        max_length: Padding length.
        truncation: Whether to truncate long inputs to `max_length`.
            If False, the long inputs are preserved as is even if they exceed
            `max_length`. Only has effect if `max_length` is specified.
        label_ignore_index:  If set, then label values of tokens that shouldn't
            contribute to the loss computation will be replaced by this special value.
        max_variable_sized_dims: Maximum number of variable-sized dimensions.
            Normally, it's 1 (sequence length dimension), but can sometimes be higher
            e.g., 2 for "cross_attention_mask" for VLM-s with multi-image inputs.
            Negative value mean `Unlimited`.
        debug: Whether to log a debug example.
        """
        self._max_length: Optional[int] = (
            int(max_length) if max_length is not None and max_length > 0 else None
        )
        self._truncation: bool = bool(truncation)

        if not hasattr(tokenizer, "padding_side") or not tokenizer.padding_side:
            raise RuntimeError("Tokenizer doesn't define `padding_side`.")
        self._padding_side = str(tokenizer.padding_side)

        if not hasattr(tokenizer, "pad_token_id") or tokenizer.pad_token_id is None:
            raise RuntimeError("Tokenizer doesn't define `pad_token_id`.")
        elif not isinstance(tokenizer.pad_token_id, int):
            raise RuntimeError(
                "Tokenizer's `pad_token_id` is not an integer. "
                f"{tokenizer.pad_token_id}. Type: {type(tokenizer.pad_token_id)}"
            )

        self._special_tokens: _SpecialTokens = _SpecialTokens(
            pad_token_id=int(tokenizer.pad_token_id),
            label_ignore_index=label_ignore_index,
        )

        self._max_input_ids_length: int = 0
        self._max_previously_logged_input_ids_length: int = 0
        self._max_variable_sized_dims: int = max_variable_sized_dims
        self._debug: bool = debug
        # Track if we've already logged an example
        self._has_logged_example: bool = False
        self._tokenizer = tokenizer  # Store tokenizer for debugging

    def _collate_simple(
        self,
        inputs_dict: dict[str, list[Any]],
        *,
        batch_max_length: int,
        padding_value_overrides: dict[str, int],
    ) -> dict[str, Any]:
        result: dict[str, Any] = {}
        for key, sequences_list in inputs_dict.items():
            try:
                padding_value = padding_value_overrides.get(key, 0)
                if self._max_variable_sized_dims == 1:
                    collated_tensor = pad_sequences(
                        sequences_list,
                        padding_side=self._padding_side,
                        padding_value=padding_value,
                    )
                else:
                    collated_tensor = pad_to_max_dim_and_stack(
                        sequences_list,
                        max_variable_sized_dims=self._max_variable_sized_dims,
                        padding_side=self._padding_side,
                        padding_value=padding_value,
                    )
                result[key] = collated_tensor
            except Exception:
                logger.error(
                    f"Failed to collate '{key}'!  "
                    f"Max variable size dims: {self._max_variable_sized_dims}, "
                    f"Batch maximum length: {batch_max_length}, "
                    f"Maximum allowed length: {self._max_length}, "
                    f"Truncation: {self._truncation}."
                )
                raise
        return result

    def __call__(self, batch) -> dict[str, Any]:
        """Pads to the longest length present in the batch.

        Args:
            batch: List of batch items.

        Returns:
            Dict[str, torch.Tensor]: Processed batch.
        """
        collation_inputs: dict[str, list[Any]] = collections.defaultdict(list)
        labels_on = _LABELS_KEY in batch[0]
        attention_mask_on = _ATTENTION_MASK_KEY in batch[0]
        cross_attention_mask_on = _CROSS_ATTENTION_MASK_KEY in batch[0]

        # Maximum sequence lengths in this batch.
        batch_max_input_ids_length: int = 0

        for item in batch:
            if _INPUT_IDS_KEY not in item:
                raise ValueError(
                    f"Item doesn't contain '{_INPUT_IDS_KEY}' key. "
                    f"Available keys: {item.keys()}"
                )
            batch_max_input_ids_length = max(
                batch_max_input_ids_length, len(item[_INPUT_IDS_KEY])
            )

            collation_inputs[_INPUT_IDS_KEY].append(item[_INPUT_IDS_KEY])
            collation_inputs[_ATTENTION_MASK_KEY].append(
                item[_ATTENTION_MASK_KEY]
                if attention_mask_on
                else create_ones_like(item[_INPUT_IDS_KEY])
            )

            if cross_attention_mask_on:
                collation_inputs[_CROSS_ATTENTION_MASK_KEY].append(
                    item[_CROSS_ATTENTION_MASK_KEY]
                )
            if labels_on:
                collation_inputs[_LABELS_KEY].append(item[_LABELS_KEY])

            if self._max_length is not None:
                if self._truncation:
                    for key in collation_inputs:
                        collation_inputs[key] = [
                            item[0 : self._max_length] for item in collation_inputs[key]
                        ]
                else:
                    for key in collation_inputs:
                        for item in collation_inputs[key]:
                            seq_len = len(item)
                            if seq_len > self._max_length:
                                raise ValueError(
                                    "Maximum sequence length exceeded. "
                                    "You should probably activate truncation. "
                                    f"'{key}' length: ({seq_len}). "
                                    f"Maximum model length: ({self._max_length})"
                                )

        # Update global (dataset) maximum lengths, and log a warning
        # about truncation if needed.
        self._update_max_lengths_and_log(
            max_input_ids_length=batch_max_input_ids_length
        )

        # Collate batch prompts.
        pad_token_id = self._special_tokens.pad_token_id
        collated_text_inputs = self._collate_simple(
            collation_inputs,
            batch_max_length=batch_max_input_ids_length,
            padding_value_overrides={
                _INPUT_IDS_KEY: pad_token_id,
                _LABELS_KEY: (
                    self._special_tokens.label_ignore_index
                    if self._special_tokens.label_ignore_index is not None
                    else pad_token_id
                ),
            },
        )

        # Combine all inputs.
        combined_batch = {
            _INPUT_IDS_KEY: collated_text_inputs[_INPUT_IDS_KEY],
            _ATTENTION_MASK_KEY: collated_text_inputs.get(_ATTENTION_MASK_KEY),
        }

        if cross_attention_mask_on:
            combined_batch[_CROSS_ATTENTION_MASK_KEY] = collated_text_inputs[
                _CROSS_ATTENTION_MASK_KEY
            ]

        # Add labels if present.
        if labels_on:
            combined_batch[_LABELS_KEY] = collated_text_inputs[_LABELS_KEY]

        # If debug is on and we haven't logged an example yet, log the first example
        if self._debug and not self._has_logged_example and len(batch) > 0:
            # Log an example of the data in the first step for debugging purposes.
            self._log_debug_example(batch, combined_batch)

        return combined_batch

    def _log_debug_example(
        self,
        batch: list[dict[str, Any]],
        combined_batch: dict[str, Any],
    ) -> None:
        """Logs a debug example if debug is enabled.

        Args:
            batch: The original batch of data.
            combined_batch: The collated batch after processing.
        """
        first_input_ids = combined_batch[_INPUT_IDS_KEY][0]
        formatted_example = self._tokenizer.decode(
            first_input_ids, skip_special_tokens=False
        )
        # Decode raw text without special tokens for raw example
        raw_text = self._tokenizer.decode(first_input_ids, skip_special_tokens=True)

        tokenized_example = []
        for tid in first_input_ids:
            if hasattr(tid, "item"):
                token_id = int(tid.item())
                decoded_token = self._tokenizer.decode([tid])
            else:
                token_id = int(tid)
                decoded_token = self._tokenizer.decode(tid)
            tokenized_example.append((token_id, decoded_token))

        model_input = {
            "input_ids": (
                first_input_ids.tolist()
                if hasattr(first_input_ids, "tolist")
                else first_input_ids
            ),
            "attention_mask": (
                combined_batch[_ATTENTION_MASK_KEY][0].tolist()
                if hasattr(combined_batch[_ATTENTION_MASK_KEY][0], "tolist")
                else combined_batch[_ATTENTION_MASK_KEY][0]
            ),
        }

        if _LABELS_KEY in combined_batch:
            lbl = combined_batch[_LABELS_KEY][0]
            model_input["labels"] = lbl.tolist() if hasattr(lbl, "tolist") else lbl

        # Mark that we've logged an example to avoid logging again
        self._has_logged_example = True
        log_example_for_debugging(
            raw_example=raw_text,
            formatted_example=str(formatted_example),
            tokenized_example=tokenized_example,
            model_input=model_input,
        )

    def _update_max_lengths_and_log(self, *, max_input_ids_length: int):
        """Updates max length counters.

        Also, logs a truncation warning if increment is large enough.
        """
        _LOG_REL_INCREMENT = 0.1  # log if max length is up 10%
        log_max_lengths: bool = False

        if max_input_ids_length > self._max_input_ids_length:
            if self._max_length is not None and max_input_ids_length > self._max_length:
                if (
                    max_input_ids_length - self._max_previously_logged_input_ids_length
                ) >= _LOG_REL_INCREMENT * self._max_previously_logged_input_ids_length:
                    log_max_lengths = True
                    self._max_previously_logged_input_ids_length = max_input_ids_length
            self._max_input_ids_length = max_input_ids_length

        if log_max_lengths:
            logger.warning(
                "Input sequence exceeded max length"
                + (" and truncated! " if self._truncation else ". ")
                + (
                    f"Max allowed length: {self._max_length}. "
                    f"'input_ids' length: {self._max_input_ids_length}."
                )
            )



import pytest

from oumi.datasets.grpo.rewards import countdown_reward


@pytest.mark.parametrize(
    "s,nums,target,reward",
    [
        # No valid answer
        ("foo bar 1", [], 1, 0),
        # Valid answer format, incorrect numbers
        ("<answer>1 + 2</answer>", [1, 3], 2, 0),
        ("<answer>1 / 2</answer>", [1, 2, 3], 6, 0),
        # Invalid equation
        ("<answer></answer>", [], 1, 0),
        ("<answer>1 foo 2 bar 3</answer>", [1, 2, 3], 1, 0),
        ("<answer>1.0 * 2.0 * 3.0</answer>", [1, 2, 3], 1, 0),
        # Incorrect answer
        ("<answer>1 + 2 + 3</answer>", [1, 2, 3], 1, 0),
        ("<answer> (1 * 2) / 3</answer>", [1, 2, 3], 1, 0),
        # Correct answer
        ("<answer> ( 3 - 2 ) * 1 </answer>", [1, 2, 3], 1, 1),
        ("<answer>(3-2)*1</answer>", [1, 2, 3], 1, 1),
    ],
)
def test_compute_soft_target_token_length_reward(s, nums, target, reward):
    ground_truth = {"target": target, "numbers": nums}
    assert countdown_reward("data_source", s, ground_truth, {}) == reward



from oumi.core.configs.params.judge_params import (
    JudgeOutputType,
    JudgeParams,
    JudgeResponseFormat,
)


def test_template_variables_applied_during_init():
    template_variables = {"role": "classifier", "domain": "medical"}

    prompt_template = "You are a {role} in the {domain} domain. Rate: {question}"
    system_instruction = "You are a {domain} expert specializing in {domain} domain."
    expected_prompt = "You are a classifier in the medical domain. Rate: {question}"
    expected_system = "You are a medical expert specializing in medical domain."

    params = JudgeParams(
        prompt_template=prompt_template,
        system_instruction=system_instruction,
        template_variables=template_variables,
        response_format=JudgeResponseFormat.XML,
        judgment_type=JudgeOutputType.BOOL,
    )
    params.replace_template_variables()

    assert params.prompt_template == expected_prompt
    assert params.system_instruction == expected_system


def test_no_template_variables_prompts_unchanged():
    template_variables = {"unused": "unused"}

    prompt_template = "Rate: {question}"
    system_instruction = "You are a helpful assistant."

    params = JudgeParams(
        prompt_template=prompt_template,
        system_instruction=system_instruction,
        template_variables=template_variables,
        response_format=JudgeResponseFormat.XML,
        judgment_type=JudgeOutputType.BOOL,
    )
    params.replace_template_variables()

    assert params.prompt_template == prompt_template
    assert params.system_instruction == system_instruction



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oumi.core.datasets import BasePretrainingDataset
from oumi.core.registry import register_dataset


@register_dataset("EleutherAI/pile")
class PileV1Dataset(BasePretrainingDataset):
    """The Pile: An 825 GiB diverse, open source language modeling dataset.

    The Pile is a large-scale English language dataset consisting of 22 smaller,
    high-quality datasets combined together. It is designed for training large
    language models and supports various natural language processing tasks
    :footcite:`2020_thepile,2022_thepile_datasheet`.

    Data Fields:
      text (str): The main text content.
      meta (dict): Metadata about the instance, including 'pile_set_name'.

    Key Features:
        - 825 GiB of diverse text data
        - Primarily in English
        - Supports text generation and fill-mask tasks
        - Includes various subsets like enron_emails, europarl, free_law, etc.

    Subsets:
      - all
      - enron_emails
      - europarl
      - free_law
      - hacker_news
      - nih_exporter
      - pubmed
      - pubmed_central
      - ubuntu_irc
      - uspto
      - github

    Splits:
      - train
      - validation
      - test

    See Also:
      - Homepage: https://pile.eleuther.ai/
      - HuggingFace hub: https://huggingface.co/datasets/EleutherAI/pile

    Warning:
        This dataset contains text from various sources and may include
        personal or sensitive information. Users should consider potential biases
        and limitations when using this dataset.

    Citations:
        .. footbibliography::
    """

    default_dataset = "EleutherAI/pile"



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base dataset class for KTO (Kahneman-Tversky Optimization).

This module provides a base class for datasets used in KTO training.
Unlike DPO which requires preference pairs, KTO works with simple binary feedback
indicating whether an output is desirable or undesirable.
"""

from collections.abc import Iterable
from typing import Any, Optional

import datasets
from typing_extensions import override

from oumi.core.datasets.base_map_dataset import BaseMapDataset, _InferredFeatureMap

_PROMPT_KEY = "prompt"
_COMPLETION_KEY = "completion"
_LABEL_KEY = "label"  # True for desirable, False for undesirable

_ROLE = "role"
_CONTENT = "content"
_ASSISTANT = "assistant"


class BaseExperimentalKtoDataset(BaseMapDataset):
    """Base class for KTO (Kahneman-Tversky Optimization) datasets.

    This class provides a comprehensive foundation for creating KTO datasets that work
    with binary feedback signals rather than preference pairs. KTO is an alignment
    method that optimizes language models based on simple binary labels indicating
    whether outputs are desirable or undesirable, making it simpler than preference-
    based methods like DPO which require paired comparisons.

    The class handles the standardization of diverse dataset formats into the
    consistent KTO format required by training frameworks. It supports both
    string-based completions and chat-formatted conversations, automatically
    extracting assistant responses when needed.

    Key Features:
        - Standardized KTO format with prompt, completion, and binary label
        - Automatic handling of chat format vs string format completions
        - Optimized feature schema for efficient dataset processing
        - Memory-efficient processing for large datasets
        - Consistent API across different KTO dataset implementations

    Dataset Format:
        The standardized KTO format includes:
        - prompt (str): The input text given to the model
        - completion (str): The model's response to be evaluated
        - label (bool): True for desirable responses, False for undesirable ones

    Usage:
        Subclasses should implement the `_load_data()` method to load their specific
        dataset format and optionally override `_transform_kto_example()` for custom
        preprocessing logic.

    Warning:
        This class is experimental and subject to change as KTO training methods
        evolve and mature.

    See Also:
        - TRL KTO Trainer: https://huggingface.co/docs/trl/main/en/kto_trainer
        - KTO Paper: https://arxiv.org/abs/2402.01306
    """

    def __init__(
        self,
        *,
        dataset_name: Optional[str] = None,
        dataset_path: Optional[str] = None,
        split: Optional[str] = None,
        **kwargs,
    ) -> None:
        """Initialize a new KTO dataset instance.

        Creates a new KTO dataset that will load and transform data into the
        standardized KTO format required for training. The constructor handles
        dataset discovery, loading, and initial setup but defers actual data
        loading until needed.

        Args:
            dataset_name (Optional[str]): Name of the dataset to load from Hugging Face
                Hub or a custom identifier. If None, uses the class's default_dataset.
            dataset_path (Optional[str]): Local path to dataset files. Takes precedence
                over dataset_name if provided. Supports .jsonl, .parquet, and cached
                HF dataset formats.
            split (Optional[str]): Dataset split to use (e.g., 'train', 'test',
                'validation'). If None and multiple splits exist, an error will be
                raised.
            **kwargs: Additional keyword arguments passed to the parent BaseMapDataset
                constructor for extended configuration options.

        Raises:
            ValueError: If neither dataset_name nor default_dataset is provided.
            FileNotFoundError: If dataset_path is specified but the file doesn't exist.

        Note:
            The actual dataset loading is deferred until the first access to ensure
            efficient initialization and allow for lazy loading patterns.
        """
        super().__init__(
            dataset_name=dataset_name,
            dataset_path=dataset_path,
            split=split,
            **kwargs,
        )

        self._data = self._load_data()

    def _transform_kto_example(self, sample: dict) -> dict:
        """Validate and transform a sample to the KTO format.

        This method processes raw dataset samples and converts them into the
        standardized format required for KTO training. Unlike DPO which requires
        preference pairs, KTO works with simple binary feedback indicating whether
        an output is desirable or undesirable.

        The method extracts the prompt, completion, and binary label from the
        input sample, handling both string and chat format completions. Chat format
        completions are processed to extract the assistant's response.

        Args:
            sample (dict): A dictionary containing the raw sample data with keys:
                - prompt: The input prompt text
                - completion: The model's response (string or chat format list)
                - label: Boolean indicating if the response is desirable (True) or
                  undesirable (False)

        Returns:
            dict: A dictionary with the standardized KTO format expected by TRL:
                - prompt (str): The input prompt text
                - completion (str): The model's response text
                - label (bool): Binary feedback - True for desirable responses,
                  False for undesirable responses

        Raises:
            ValueError: If the completion is in chat format but no assistant turn
                is found.
        """
        prompt = sample[_PROMPT_KEY]
        completion = sample[_COMPLETION_KEY]
        label = sample[_LABEL_KEY]

        # Extract text from completion if it's in chat format
        if isinstance(completion, list):
            completion = self._extract_from_chat_format(completion)

        return {
            _PROMPT_KEY: prompt,
            _COMPLETION_KEY: completion,
            _LABEL_KEY: label,
        }

    @override
    def transform(self, sample: dict) -> dict:
        """Transform a raw dataset sample into the standardized format.

        This is the main entry point for processing dataset samples. It delegates
        to the KTO-specific transformation method to ensure consistent formatting
        across all KTO datasets.

        Args:
            sample (dict): A raw dataset sample containing prompt, completion,
                and label information.

        Returns:
            dict: The transformed sample in KTO format with standardized keys:
                prompt, completion, and label.
        """
        return self._transform_kto_example(sample)

    def _extract_from_chat_format(self, sample) -> str:
        """Extract the assistant's response from a chat-formatted completion.

        This method handles completions that are provided in chat format (as a list
        of message dictionaries) by extracting the content from the last assistant
        turn. If the sample is already a string, it returns it unchanged.

        The method searches backwards through the conversation to find the most recent
        assistant response, which is typically the completion we want to evaluate.

        Args:
            sample: The completion data, either as a string or a list of message
                dictionaries in chat format. Each message dict should have 'role'
                and 'content' keys.

        Returns:
            str: The extracted assistant response content as a string.

        Raises:
            ValueError: If the sample is in chat format (list) but no message
                with role 'assistant' is found.

        Example:
            >>> chat_format = [
            ...     {"role": "user", "content": "Hello"},
            ...     {"role": "assistant", "content": "Hi there!"}
            ... ]
            >>> self._extract_from_chat_format(chat_format)
            'Hi there!'
        """
        if not isinstance(sample, list):
            return sample

        for turn in sample[::-1]:
            if turn[_ROLE] == _ASSISTANT:
                return turn[_CONTENT]

        raise ValueError("No chat turn was found with an 'assistant' role.")

    @property
    def _kto_features(self) -> datasets.Features:
        """Get the explicit feature schema definition for KTO training datasets.

        This property defines the standardized schema that all KTO datasets must
        conform to. The schema ensures type safety and consistency across different
        KTO dataset implementations and is used by the Hugging Face datasets library
        for efficient serialization and validation.

        The KTO format requires exactly three fields:
        - prompt: The input text that was given to the model
        - completion: The model's generated response to be evaluated
        - label: Binary feedback indicating if the completion is desirable

        Returns:
            datasets.Features: A Features object defining the schema with:
                - prompt (string): Input prompt text
                - completion (string): Model response text
                - label (bool): True for desirable responses, False for undesirable
        """
        return datasets.Features(
            {
                "prompt": datasets.Value("string"),
                "completion": datasets.Value("string"),
                "label": datasets.Value("bool"),
            }
        )

    @override
    def _detect_features_and_estimate_element_size_bytes(
        self, samples_iter: Iterable[dict[str, Any]]
    ) -> _InferredFeatureMap:
        """Detect dataset features and estimate memory requirements for KTO datasets.

        This method overrides the base implementation to provide KTO-specific feature
        detection and memory estimation. Instead of inferring features from sample data,
        it uses the predefined KTO schema to ensure consistency and optimize for the
        specific structure of KTO training data.

        The method samples a subset of the dataset to estimate average element size,
        which is used for memory management and efficient dataset processing. The
        feature map is marked as optimized since we use explicit KTO features rather
        than inferred ones.

        Args:
            samples_iter (Iterable[dict[str, Any]]): An iterable of dataset samples
                to analyze for feature detection and size estimation.

        Returns:
            _InferredFeatureMap: A named tuple containing:
                - feature_map: The explicit KTO features schema
                - is_feature_map_optimized: Always True for KTO datasets
                - element_size_in_bytes: Estimated average size per sample
                - multimodal: Always False for text-only KTO data
        """
        from oumi.utils.torch_utils import estimate_sample_dict_size_in_bytes

        # Collect a few samples to estimate average size
        samples = []
        samples_iterator = iter(samples_iter)
        for _ in range(min(10, len(self))):  # Use up to 10 samples
            try:
                samples.append(next(samples_iterator))
            except StopIteration:
                break

        # Calculate estimated element size based on actual samples
        element_size = 1024  # Default fallback
        if samples:
            # Get average size of samples
            element_size = sum(
                estimate_sample_dict_size_in_bytes(s) for s in samples
            ) // len(samples)
            # Add 20% buffer for safety
            element_size = int(element_size * 1.2)

        # Return features optimized for KTO training with proper size estimate
        return _InferredFeatureMap(
            feature_map=self._kto_features,
            is_feature_map_optimized=True,
            element_size_in_bytes=element_size,
            multimodal=False,
        )



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from pathlib import Path
from typing import Any, Optional

from oumi.core.configs import EvaluationConfig, EvaluationTaskParams
from oumi.core.evaluation.evaluation_result import EvaluationResult
from oumi.utils.logging import logger
from oumi.utils.serialization_utils import json_serializer
from oumi.utils.version_utils import get_python_package_versions

# Output filenames for saving evaluation results and reproducibility information.
OUTPUT_FILENAME_TASK_RESULT = "task_result.json"
OUTPUT_FILENAME_TASK_PARAMS = "task_params.json"
OUTPUT_FILENAME_BACKEND_CONFIG = "backend_config.json"
OUTPUT_FILENAME_MODEL_PARAMS = "model_params.json"
OUTPUT_FILENAME_GENERATION_PARAMS = "generation_params.json"
OUTPUT_FILENAME_INFERENCE_PARAMS = "inference_params.json"
OUTPUT_FILENAME_PACKAGE_VERSIONS = "package_versions.json"
OUTPUT_FILENAME_EVALUATION_CONFIG_YAML = "evaluation_config.yaml"


def _save_to_file(output_path: Path, data: Any) -> None:
    """Serialize and save `data` to `output_path`."""
    with open(output_path, "w") as file_out:
        file_out.write(json_serializer(data))


def _find_non_existing_output_dir_from_base_dir(base_dir: Path) -> Path:
    """Finds a new output directory, if the provided `base_dir` already exists.

    Why is this function useful?
        Users may repeatedly run the same evaluation script, which will overwrite the
        results of the existing output directory. When this happens, we could fail, to
        avoid corrupting previous evaluation results. However, for a more user-friendly
        experience, we can automatically create a new output directory with a unique
        name. This function does this by appending an index to the base directory that
        was provided (`base_dir`), as follows: `<base_dir>_<index>`.

    Args:
        base_dir: The base directory where the evaluation results would be saved.

    Returns:
        A new output directory (does not exist yet), if `base_dir` already exists,
        or the original `base_dir` if it does not exist.
    """
    dir_index = 0
    new_dir = base_dir

    while new_dir.exists():
        logger.warning(
            f"The requested output directory already exists: `{new_dir}`. Looking up a "
            "new location, to avoid overwriting previous evaluation results."
        )
        dir_index += 1
        new_dir = base_dir.parent / f"{base_dir.name}_{dir_index}"

    if dir_index > 0:
        logger.warning(
            f"Created a new output directory to avoid overwriting previous evaluation "
            f"results. The new directory is `{new_dir}`."
        )

    return new_dir


def save_evaluation_output(
    backend_name: str,
    task_params: EvaluationTaskParams,
    evaluation_result: EvaluationResult,
    base_output_dir: Optional[str],
    config: Optional[EvaluationConfig],
) -> None:
    """Writes configuration settings and evaluation outputs to files.

    Args:
        backend_name: The name of the evaluation backend used (e.g., "lm_harness").
        task_params: Oumi task parameters used for this evaluation.
        evaluation_result: The evaluation results to save.
        base_output_dir: The directory where the evaluation results will be saved.
            A subdirectory with the name `<base_output_dir> / <backend_name>_<time>`
            will be created to retain all files related to this evaluation. If there is
            an existing directory with the same name, a new directory with a unique
            index will be created: `<base_output_dir> / <backend_name>_<time>_<index>`.
        config: Oumi evaluation configuration settings used for the evaluation.
    """
    # Ensure the evaluation backend and output directory are valid.
    if not backend_name:
        raise ValueError("The evaluation backend name must be provided.")
    base_output_dir = base_output_dir or "."

    # Create the output directory: `<base_output_dir> / <backend_name>_<time>`.
    start_time_in_path = (
        f"_{evaluation_result.start_time}" if evaluation_result.start_time else ""
    )
    output_dir = Path(base_output_dir) / f"{backend_name}{start_time_in_path}"
    if output_dir.exists():
        output_dir = _find_non_existing_output_dir_from_base_dir(output_dir)
    output_dir.mkdir(parents=True, exist_ok=False)

    # Save all evaluation metrics, start date/time, and duration.
    if evaluation_result.task_result:
        task_result = copy.deepcopy(evaluation_result.task_result)
    else:
        task_result = {}
    if evaluation_result.start_time:
        task_result["start_time"] = evaluation_result.start_time
    if evaluation_result.elapsed_time_sec:
        task_result["duration_sec"] = evaluation_result.elapsed_time_sec
    if task_result:
        _save_to_file(output_dir / OUTPUT_FILENAME_TASK_RESULT, task_result)

    # Save backend-specific task configuration.
    if evaluation_result.backend_config:
        _save_to_file(
            output_dir / OUTPUT_FILENAME_BACKEND_CONFIG,
            evaluation_result.backend_config,
        )

    # Save Oumi's task parameters/configuration.
    _save_to_file(output_dir / OUTPUT_FILENAME_TASK_PARAMS, task_params)

    # Save all relevant Oumi configurations.
    if config:
        try:
            config.to_yaml(output_dir / OUTPUT_FILENAME_EVALUATION_CONFIG_YAML)
        except Exception as e:
            logger.error(f"Failed to save EvaluationConfig as YAML: {e}")

        if config.model:
            _save_to_file(output_dir / OUTPUT_FILENAME_MODEL_PARAMS, config.model)
        if config.generation:
            _save_to_file(
                output_dir / OUTPUT_FILENAME_GENERATION_PARAMS, config.generation
            )
        if config.inference_engine or config.inference_remote_params:
            inference_params = {
                "engine": config.inference_engine,
                "remote_params": config.inference_remote_params,
            }
            _save_to_file(
                output_dir / OUTPUT_FILENAME_INFERENCE_PARAMS, inference_params
            )

    # Save python environment (package versions).
    package_versions = get_python_package_versions()
    _save_to_file(output_dir / OUTPUT_FILENAME_PACKAGE_VERSIONS, package_versions)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing_extensions import override  # noqa: I001

from oumi.core.datasets import VisionLanguageSftDataset
from oumi.core.registry import register_dataset
from oumi.core.types.conversation import (
    ContentItem,
    Conversation,
    Message,
    Role,
    Type,
)


@register_dataset("allenai/pixmo-cap-qa")
class PixmoCapQADataset(VisionLanguageSftDataset):
    """Dataset class for the `allenai/pixmo-cap-qa` dataset.

    The dataset is affected by some image URLs having a 404 issue.
    """

    default_dataset = "allenai/pixmo-cap-qa"

    @override
    def transform_conversation(self, example: dict) -> Conversation:
        """Transform the example into a Conversation object.

        Sample "question": "[USER] Can you come up with a joke? [ASSISTANT]"
        It starts with a [USER] and ends with an [ASSISTANT] role tag.
        The Assistant response appears in the "answer" field.
        """
        messages: list[Message] = []
        messages.append(
            Message(
                role=Role.USER,
                content=[
                    ContentItem(type=Type.IMAGE_URL, content=example["image_url"]),
                ],
            )
        )
        questions = example["question"].strip().split("\n")
        for question in questions:
            if question.startswith("[USER]"):
                question = question[len("[USER]") :].strip()
                messages.append(Message(role=Role.USER, content=question))
            elif question.startswith("[ASSISTANT]"):
                answer = question[len("[ASSISTANT]") :].strip()
                messages.append(Message(role=Role.ASSISTANT, content=answer))
        messages.append(Message(role=Role.ASSISTANT, content=example["answer"]))
        return Conversation(messages=messages)



from pathlib import Path

import pytest

from oumi.core.types.conversation import Conversation, Message, Role
from oumi.utils.logging import get_logger


@pytest.fixture
def root_testdata_dir() -> Path:
    return Path(__file__).parent / "testdata"


@pytest.fixture(scope="session", autouse=True)
def setup_logging():
    """Fixture to set up logging for all tests.

    We want to propagate to the root logger so that
    pytest caplog can capture logs, and we can test
    logging for the default oumi logger.
    """
    logger = get_logger("oumi")
    logger.propagate = True
    return logger


@pytest.fixture(autouse=True)
def retain_logging_level():
    """Fixture to preserve the logging level between tests."""
    logger = get_logger("oumi")
    # Store the current log level
    log_level = logger.level
    yield
    # Rehydrate the log level
    logger.setLevel(log_level)


@pytest.fixture(autouse=True)
def cleanup_gpu_memory(request):
    """Automatically clean up GPU memory after GPU tests."""
    yield  # Let the test run first

    # Only cleanup for GPU-related tests to avoid overhead
    gpu_markers = {"single_gpu", "multi_gpu"}
    test_markers = {mark.name for mark in request.node.iter_markers()}

    # Also check for GPU-related decorators in the test
    has_gpu_decorator = any(
        "requires_cuda" in str(mark) or "requires_gpu" in str(mark)
        for mark in request.node.iter_markers()
    )

    # Check if test function uses GPU device mapping
    test_source = ""
    try:
        import inspect

        test_source = inspect.getsource(request.node.function)
        has_device_map = "get_default_device_map_for_inference" in test_source
    except Exception:
        has_device_map = False

    if gpu_markers.intersection(test_markers) or has_gpu_decorator or has_device_map:
        try:
            import gc

            import torch

            if torch.cuda.is_available():
                # More aggressive GPU memory cleanup
                torch.cuda.empty_cache()
                torch.cuda.reset_peak_memory_stats()
                torch.cuda.synchronize()
                gc.collect()
                # Force another cache clear after garbage collection
                torch.cuda.empty_cache()
        except Exception:
            # Silently ignore cleanup errors to avoid test failures
            pass


@pytest.fixture
def single_turn_conversation():
    return Conversation(
        messages=[
            Message(role=Role.USER, content="Hello"),
            Message(role=Role.ASSISTANT, content="Hi there!"),
        ]
    )



import argparse
import functools
import os
import pathlib
import time
from collections.abc import Iterator
from typing import (
    Any,
    NamedTuple,
    Optional,
    TypeVar,
    Union,
    cast,
)

import datasets
from tqdm import tqdm
from transformers import PreTrainedTokenizer, PreTrainedTokenizerFast

from oumi.builders import build_tokenizer
from oumi.core.configs import TrainingConfig
from oumi.utils.logging import logger

_TOKEN_IDS_COLUMN_NAME = "input_ids"  # The common convention.

DatasetType = TypeVar("DatasetType", datasets.Dataset, datasets.IterableDataset)


def _list_input_files(
    input_paths: list[str],
    input_format: str,
) -> Iterator[pathlib.Path]:
    for path_str in input_paths:
        path = pathlib.Path(path_str)
        if not path.exists():
            logger.warning(f"{path} not found and skipped")
            continue
        yield from path.glob(f"*.{input_format}") if path.is_dir() else [path]


def _tokenize_examples(
    tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast],
    target_col: str,
    examples: dict[str, Any],
) -> dict[str, Any]:
    batch = tokenizer(examples[target_col])
    token_ids: list[list[int]] = batch.input_ids
    result = examples.copy()
    result[_TOKEN_IDS_COLUMN_NAME] = token_ids
    return result


def _tokenize_dataset_impl(
    dataset: DatasetType,
    tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast],
    target_col: str,
    num_proc: int,
    keep_in_memory: bool,
) -> DatasetType:
    logger.info("Tokenizing the dataset...")
    dataset = dataset.map(
        functools.partial(_tokenize_examples, tokenizer, target_col),
        batched=True,
        batch_size=128,
        keep_in_memory=keep_in_memory,
        num_proc=num_proc,
    )
    logger.info("Finished tokenizing the dataset.")
    return dataset


def _process_file(
    tokenizer: Optional[Union[PreTrainedTokenizer, PreTrainedTokenizerFast]],
    target_col: str,
    input_file: pathlib.Path,
    input_format: str,
    output_parquet_file: pathlib.Path,
    num_proc: int,
) -> None:
    logger.info(f"Loading {input_file}.")
    if input_format == "jsonl":
        dataset = datasets.Dataset.from_json(str(input_file), keep_in_memory=True)
    elif input_format == "arrow":
        dataset = datasets.Dataset.from_file(str(input_file), in_memory=True)
    else:
        assert input_format == "parquet"
        dataset = datasets.Dataset.from_parquet(str(input_file), keep_in_memory=True)

    if tokenizer is not None:
        dataset = _tokenize_dataset_impl(
            cast(datasets.Dataset, dataset),
            tokenizer,
            target_col,
            num_proc=num_proc,
            keep_in_memory=True,
        )

    logger.info(f"Writing the processsed data to {output_parquet_file}.")
    dataset.to_parquet(output_parquet_file)
    logger.info(f"Finished writing to {output_parquet_file}.")


def _process_dataset(
    *,
    tokenizer: Optional[Union[PreTrainedTokenizer, PreTrainedTokenizerFast]],
    target_col: str,
    input_dataset: str,
    dataset_subset: Optional[str],
    dataset_split: Optional[str],
    trust_remote_code: bool,
    output_dataset_path: pathlib.Path,
    num_shards: int,
    max_shard_size: str,
    num_proc: int,
) -> None:
    if (
        input_dataset.startswith("/")
        or input_dataset.startswith(".")
        or input_dataset.startswith("~")
    ):
        input_dataset_path = pathlib.Path(input_dataset)
        logger.info(f"Loading {input_dataset_path} from disk...")
        dataset = datasets.Dataset.load_from_disk(str(input_dataset_path))
    else:
        logger.info(f"Loading {input_dataset} from HuggingFace...")
        splits_or_dataset = datasets.load_dataset(
            path=input_dataset,
            name=(dataset_subset or None),
            split=(dataset_split or None),
            trust_remote_code=trust_remote_code,
        )

        if isinstance(
            splits_or_dataset, (datasets.IterableDataset, datasets.IterableDatasetDict)
        ):
            raise ValueError("IterableDataset is not supported with this class.")

        if isinstance(splits_or_dataset, datasets.Dataset):
            dataset = splits_or_dataset
        elif dataset_split:
            dataset = splits_or_dataset[dataset_split]
        elif len(splits_or_dataset) == 1:
            dataset = splits_or_dataset.values().__iter__().__next__()
        else:
            raise ValueError(
                "Multiple splits found in the dataset. Please specify a single split. "
                f"Available splits: {list(splits_or_dataset.keys())}"
            )

    logger.info(
        "\n".join(
            [
                "Dataset Loaded!",
                f"Split: {dataset.split}",
                f"Version: {dataset.version}",
                f"Dataset size: {dataset.dataset_size}",
                f"Download size: {dataset.download_size}",
                f"Size: {dataset.size_in_bytes} bytes",
                f"Column names: {dataset.column_names}",
            ]
        )
    )

    if tokenizer is not None:
        dataset = _tokenize_dataset_impl(
            cast(datasets.Dataset, dataset),
            tokenizer,
            target_col,
            num_proc=num_proc,
            keep_in_memory=False,
        )

    logger.info(f"Writing the output dataset to {output_dataset_path} ...")
    dataset.save_to_disk(
        str(output_dataset_path),
        num_shards=(num_shards if not max_shard_size and num_shards > 0 else None),
        max_shard_size=(max_shard_size or None),
        num_proc=num_proc,
    )
    logger.info(f"Finished writing to {output_dataset_path} !")


class ParsedArgs(NamedTuple):
    config_path: str
    verbose: bool
    input_dataset: str
    dataset_subset: str
    dataset_split: str
    trust_remote_code: bool
    input_paths: list[str]
    input_format: str
    target_col: str
    output_dir: str
    num_shards: int
    max_shard_size: str
    overwrite: bool
    num_proc: int
    skip_tokenize: bool


def parse_cli() -> tuple[ParsedArgs, list[str]]:
    """Parses command line arguments and returns the configuration filename."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c",
        "--config",
        default=None,
        help="Path to the configuration file",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
    )
    parser.add_argument(
        "--target_col",
        type=str,
        default="",
        help="Target text column to tokenize.",
    )
    parser.add_argument(
        "--overwrite",
        action="store_true",
        help="Whether to overwrite the output directory.",
    )
    parser.add_argument(
        "--num_proc",
        type=int,
        default=-1,
        help=(
            "Number of processes for parallel execution. "
            "If -1, then use all available CPU cores."
        ),
    )
    parser.add_argument(
        "--skip_tokenize",
        action="store_true",
        help=(
            "Whether to skip tokenization. "
            "Can be useful if you just want to copy a dataset, "
            "or convert it from one format to another."
        ),
    )

    parser.add_argument(
        "--output_dir",
        type=str,
        required=True,
        help="Path to the output directory.",
    )
    parser.add_argument(
        "--max_shard_size",
        type=str,
        default="",
        help="Max shard size e.g., 256MB",
    )
    parser.add_argument(
        "--num_shards",
        type=int,
        default=512,
        help="Number of shards.",
    )

    # Parameters to work with HF datasets
    parser.add_argument(
        "--input_dataset",
        type=str,
        help="Directory path to the input Huggingface dataset, or a dataset name.",
    )
    parser.add_argument(
        "--dataset_subset",
        type=str,
        help="Subset of an input dataset",
    )
    parser.add_argument(
        "--dataset_split",
        type=str,
        help="Split of an input dataset e.g., 'train'",
    )
    parser.add_argument(
        "--trust_remote_code",
        action="store_true",
        help="Whether to trust remote code.",
    )

    # Parameters to work with individual files
    parser.add_argument(
        "--input_path",
        type=str,
        nargs="+",
        help="Path(s) to the input data directory or file.",
    )
    parser.add_argument(
        "--input_format",
        type=str,
        default="parquet",
        choices=["jsonl", "parquet", "arrow"],
        help="Input format.",
    )

    args, unknown = parser.parse_known_args()
    return (
        ParsedArgs(
            config_path=args.config,
            verbose=args.verbose,
            input_dataset=args.input_dataset,
            dataset_subset=args.dataset_subset,
            dataset_split=args.dataset_split,
            trust_remote_code=args.trust_remote_code,
            input_paths=args.input_path,
            input_format=args.input_format,
            target_col=args.target_col,
            output_dir=args.output_dir,
            num_shards=args.num_shards,
            max_shard_size=args.max_shard_size,
            overwrite=args.overwrite,
            num_proc=args.num_proc,
            skip_tokenize=args.skip_tokenize,
        ),
        unknown,
    )


def main() -> None:
    """Main function."""
    parsed_args, arg_list = parse_cli()

    logger.info(f"Parsed arguments: {parsed_args}")
    logger.info(f"Unknown arguments: {arg_list}")

    config: TrainingConfig = None
    target_col: str = ""
    tokenizer: Optional[Union[PreTrainedTokenizer, PreTrainedTokenizerFast]] = None
    if not parsed_args.skip_tokenize:
        config = TrainingConfig.from_yaml_and_arg_list(
            parsed_args.config_path, arg_list, logger=logger
        )

        # Find first non-empty value as target column name.
        target_col = next(
            s
            for s in [
                parsed_args.target_col,
                config.data.train.target_col,
                config.data.validation.target_col,
                config.data.test.target_col,
                "text",
            ]
            if s
        )
        logger.info("Initializing the tokenizer...")
        tokenizer = build_tokenizer(config.model)

    output_dir: pathlib.Path = pathlib.Path(parsed_args.output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    start_time = time.time()

    num_proc = (
        (os.cpu_count() or 1) if parsed_args.num_proc == -1 else parsed_args.num_proc
    )

    if parsed_args.input_dataset:
        logger.info(f"Processing the dataset {parsed_args.input_dataset}...")
        _process_dataset(
            tokenizer=tokenizer,
            target_col=target_col,
            input_dataset=parsed_args.input_dataset,
            dataset_subset=parsed_args.dataset_subset,
            dataset_split=parsed_args.dataset_split,
            trust_remote_code=parsed_args.trust_remote_code,
            output_dataset_path=output_dir,
            num_shards=parsed_args.num_shards,
            max_shard_size=parsed_args.max_shard_size,
            num_proc=num_proc,
        )
    else:
        datasets.disable_caching()
        input_files: list[pathlib.Path] = list(
            sorted(_list_input_files(parsed_args.input_paths, parsed_args.input_format))
        )
        if not input_files:
            logger.warning("No files found!")
            return

        logger.info(f"Loading the dataset from {len(input_files)} files...")
        for input_file in tqdm(input_files):
            output_file: pathlib.Path = output_dir / f"{input_file.stem}.parquet"

            if output_file.exists() and not parsed_args.overwrite:
                logger.error(f"{output_file} already exists. Specify --overwrite.")
                continue
            _process_file(
                tokenizer,
                target_col,
                input_file,
                parsed_args.input_format,
                output_file,
                num_proc=num_proc,
            )

    end_time = time.time()
    logger.info(
        f"Finished processing the dataset. Elapsed time: {end_time - start_time} sec!"
    )


if __name__ == "__main__":
    main()



import dataclasses
from enum import Enum
from typing import Any, Optional

import torch

from oumi.core.configs.params.model_params import ModelParams
from oumi.core.configs.params.training_params import TrainingParams
from oumi.core.configs.training_config import TrainingConfig
from oumi.utils.serialization_utils import flatten_config


class ConfigTestEnum(Enum):
    VALUE_A = "a"
    VALUE_B = "b"


@dataclasses.dataclass
class SimpleConfig:
    name: str = "test"
    value: int = 42
    enabled: bool = True


@dataclasses.dataclass
class NestedConfig:
    simple: SimpleConfig = dataclasses.field(default_factory=SimpleConfig)
    settings: dict[str, Any] = dataclasses.field(default_factory=dict)


@dataclasses.dataclass
class ComplexConfig:
    name: str = "complex"
    nested: NestedConfig = dataclasses.field(default_factory=NestedConfig)
    items: list[str] = dataclasses.field(default_factory=list)
    optional_value: Optional[int] = None
    enum_value: ConfigTestEnum = ConfigTestEnum.VALUE_A
    torch_dtype: torch.dtype = torch.float32


class TestFlattenConfig:
    def test_flatten_simple_dict(self):
        """Test flattening a simple dictionary."""
        config = {"name": "test", "value": 42, "enabled": True}
        result = flatten_config(config)

        expected = {"name": "test", "value": 42, "enabled": True}
        assert result == expected

    def test_flatten_nested_dict(self):
        """Test flattening a nested dictionary."""
        config = {
            "name": "test",
            "settings": {"debug": True, "level": 2, "nested": {"deep": "value"}},
        }
        result = flatten_config(config)

        expected = {
            "name": "test",
            "settings.debug": True,
            "settings.level": 2,
            "settings.nested.deep": "value",
        }
        assert result == expected

    def test_flatten_simple_dataclass(self):
        """Test flattening a simple dataclass."""
        config = SimpleConfig(name="test_name", value=100, enabled=False)
        result = flatten_config(config)

        expected = {"name": "test_name", "value": 100, "enabled": False}
        assert result == expected

    def test_flatten_nested_dataclass(self):
        """Test flattening nested dataclasses."""
        simple = SimpleConfig(name="nested_test", value=200, enabled=True)
        config = NestedConfig(simple=simple, settings={"key": "value"})
        result = flatten_config(config)

        expected = {
            "simple.name": "nested_test",
            "simple.value": 200,
            "simple.enabled": True,
            "settings.key": "value",
        }
        assert result == expected

    def test_flatten_complex_dataclass(self):
        """Test flattening a complex dataclass with various types."""
        config = ComplexConfig(
            name="complex_test",
            nested=NestedConfig(
                simple=SimpleConfig(name="inner", value=300, enabled=False),
                settings={"mode": "production", "timeout": 30},
            ),
            items=["item1", "item2", "item3"],
            optional_value=42,
            enum_value=ConfigTestEnum.VALUE_B,
            torch_dtype=torch.float16,
        )
        result = flatten_config(config)

        expected = {
            "name": "complex_test",
            "nested.simple.name": "inner",
            "nested.simple.value": 300,
            "nested.simple.enabled": False,
            "nested.settings.mode": "production",
            "nested.settings.timeout": 30,
            "items": "['item1', 'item2', 'item3']",
            "optional_value": 42,
            "enum_value": "ConfigTestEnum.VALUE_B",
            "torch_dtype": "torch.float16",
        }
        assert result == expected

    def test_flatten_with_custom_options(self):
        """Test flattening with custom prefix and separator."""
        config = {"settings": {"debug": True, "level": 2}}

        # Test with prefix
        result = flatten_config({"name": "test"}, prefix="config")
        assert result == {"config.name": "test"}

        # Test with custom separator
        result = flatten_config(config, separator="_")
        assert result == {"settings_debug": True, "settings_level": 2}

        # Test with both prefix and separator
        result = flatten_config({"debug": True}, prefix="app", separator="__")
        assert result == {"app__debug": True}

    def test_flatten_empty_dict(self):
        """Test flattening an empty dictionary."""
        config = {}
        result = flatten_config(config)

        expected = {}
        assert result == expected

    def test_flatten_none_values(self):
        """Test flattening with None values."""
        config = {"name": "test", "value": None, "settings": {"debug": None}}
        result = flatten_config(config)

        expected = {"name": "test", "value": None, "settings.debug": None}
        assert result == expected

    def test_flatten_list_with_simple_values(self):
        """Test flattening lists with simple values."""
        config = {"items": [1, 2, 3], "names": ["alice", "bob"]}
        result = flatten_config(config)

        expected = {"items": "[1, 2, 3]", "names": "['alice', 'bob']"}
        assert result == expected

    def test_flatten_list_with_dicts(self):
        """Test flattening lists containing dictionaries."""
        config = {"users": [{"name": "alice", "age": 30}, {"name": "bob", "age": 25}]}
        result = flatten_config(config)

        expected = {
            "users.0.name": "alice",
            "users.0.age": 30,
            "users.1.name": "bob",
            "users.1.age": 25,
        }
        assert result == expected

    def test_flatten_list_with_dataclasses(self):
        """Test flattening lists containing dataclasses."""
        config = {
            "configs": [
                SimpleConfig(name="config1", value=10, enabled=True),
                SimpleConfig(name="config2", value=20, enabled=False),
            ]
        }
        result = flatten_config(config)

        expected = {
            "configs.0.name": "config1",
            "configs.0.value": 10,
            "configs.0.enabled": True,
            "configs.1.name": "config2",
            "configs.1.value": 20,
            "configs.1.enabled": False,
        }
        assert result == expected

    def test_flatten_tuple(self):
        """Test flattening tuples."""
        config = {"coordinates": (10, 20, 30), "pair": ("x", "y")}
        result = flatten_config(config)

        expected = {"coordinates": "(10, 20, 30)", "pair": "('x', 'y')"}
        assert result == expected

    def test_flatten_non_dict_objects(self):
        """Test flattening non-dict, non-dataclass objects."""
        # Test simple string
        result = flatten_config("simple_string")
        assert result == {"value": "simple_string"}

        # Test with prefix
        result = flatten_config(42, prefix="number")
        assert result == {"number": "42"}

    def test_flatten_special_types(self):
        """Test flattening with special types like torch.dtype."""
        config = {
            "dtype": torch.float32,
            "device": torch.device("cpu"),
            "tensor_shape": torch.Size([2, 3, 4]),
        }
        result = flatten_config(config)

        expected = {
            "dtype": "torch.float32",
            "device": "cpu",
            "tensor_shape": "torch.Size([2, 3, 4])",
        }
        assert result == expected

    def test_flatten_mixed_types(self):
        """Test flattening with mixed types including complex objects."""
        config = {
            "string": "test",
            "integer": 42,
            "float": 3.14,
            "boolean": True,
            "none_value": None,
            "list": [1, 2, 3],
            "dict": {"nested": "value"},
            "enum": ConfigTestEnum.VALUE_A,
            "torch_dtype": torch.float16,
        }
        result = flatten_config(config)

        expected = {
            "string": "test",
            "integer": 42,
            "float": 3.14,
            "boolean": True,
            "none_value": None,
            "list": "[1, 2, 3]",
            "dict.nested": "value",
            "enum": "ConfigTestEnum.VALUE_A",
            "torch_dtype": "torch.float16",
        }
        assert result == expected

    def test_flatten_deeply_nested_structure(self):
        """Test flattening deeply nested structures."""
        config = {"level1": {"level2": {"level3": {"level4": {"value": "deep_value"}}}}}
        result = flatten_config(config)

        expected = {"level1.level2.level3.level4.value": "deep_value"}
        assert result == expected

    def test_flatten_empty_containers(self):
        """Test flattening empty containers."""
        config = {"empty_list": [], "empty_dict": {}}
        result = flatten_config(config)

        assert result == {"empty_list": "[]"}

    def test_flatten_preserves_basic_types(self):
        """Test that basic types are preserved without conversion."""
        config = {
            "string": "test",
            "int": 42,
            "float": 3.14,
            "bool": True,
            "none": None,
        }
        result = flatten_config(config)

        # Check that basic types are preserved
        assert isinstance(result["string"], str)
        assert isinstance(result["int"], int)
        assert isinstance(result["float"], float)
        assert isinstance(result["bool"], bool)
        assert result["none"] is None

    def test_flatten_training_config(self):
        """Test flattening with actual oumi TrainingConfig."""
        config = TrainingConfig(
            model=ModelParams(model_name="microsoft/DialoGPT-medium"),
            training=TrainingParams(
                output_dir="./outputs",
                learning_rate=3e-5,
                num_train_epochs=3,
                per_device_train_batch_size=2,
            ),
        )
        result = flatten_config(config)

        # Should have flattened all nested parameters
        assert "model.model_name" in result
        assert "training.output_dir" in result
        assert "training.learning_rate" in result
        assert "training.num_train_epochs" in result
        assert len(result) > 10  # Should have many flattened parameters

    def test_flatten_enum_values(self):
        """Test that enum values are extracted correctly."""
        config = {"enum_a": ConfigTestEnum.VALUE_A, "enum_b": ConfigTestEnum.VALUE_B}
        result = flatten_config(config)

        assert result["enum_a"] == "ConfigTestEnum.VALUE_A"
        assert result["enum_b"] == "ConfigTestEnum.VALUE_B"



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import re

from oumi.core.registry import RegistryType, register


def _whitespace_tokenize(s: str) -> list[str]:
    return re.split(r"\s+", s)


def compute_soft_target_token_length_reward(num_tokens: int, *, target_tokens: int):
    """Returns maximum reward for inputs that are `target_tokens` long.

    The reward is in the [0,1] range and reduces smoothly from the maximum value of 1.0
    if the actual number of tokens deviates from `target_tokens`.

    The reward is proportional to: `x*exp(-x)` where `x := num_tokens/target_tokens`.
    """
    x = num_tokens / target_tokens
    return x * math.exp(-x) * math.e


def _compute_completion_soft_target_token_length_reward(
    completions: list[str], *, target_tokens: int
):
    return [
        compute_soft_target_token_length_reward(
            len(_whitespace_tokenize(content)), target_tokens=target_tokens
        )
        for content in completions
    ]


def compute_sharp_target_token_length_reward(num_tokens: int, *, target_tokens: int):
    """Returns maximum reward for inputs that are `target_tokens` long.

    The reward reduces sharply if the actual number of tokens deviates
    from `target_tokens`.

    The reward is computed as: `-|num_tokens - target_tokens|`, which penalizes
    token counts not equal to `target_tokens`.
    """
    return -abs(num_tokens - target_tokens)


def _compute_completion_sharp_target_token_length_reward(
    completions: list[str], *, target_tokens: int
):
    return [
        compute_sharp_target_token_length_reward(
            len(_whitespace_tokenize(content)), target_tokens=target_tokens
        )
        for content in completions
    ]


# Simple toy length-based reward functions for experimentation and demonstration
# purposes. In practice, most users are expected to define and  use custom reward
# functions, not these.
# For more details on custom reward functions used in trl's GRPOTrainer, see:
# https://huggingface.co/docs/trl/main/en/grpo_trainer#using-a-custom-reward-function.


@register("soft_5tokens_completions", RegistryType.REWARD_FUNCTION)
def _soft_5tokens_completions(completions: list[str], **kwargs):
    return _compute_completion_soft_target_token_length_reward(
        completions, target_tokens=5
    )


@register("soft_10tokens_completions", RegistryType.REWARD_FUNCTION)
def _soft_10tokens_completions(completions: list[str], **kwargs):
    return _compute_completion_soft_target_token_length_reward(
        completions, target_tokens=10
    )


@register("soft_20tokens_completions", RegistryType.REWARD_FUNCTION)
def _soft_20tokens_completions(completions: list[str], **kwargs):
    return _compute_completion_soft_target_token_length_reward(
        completions, target_tokens=20
    )


@register("sharp_5tokens_completions", RegistryType.REWARD_FUNCTION)
def _sharp_5tokens_completions(completions: list[str], **kwargs):
    return _compute_completion_sharp_target_token_length_reward(
        completions, target_tokens=5
    )


@register("sharp_10tokens_completions", RegistryType.REWARD_FUNCTION)
def _sharp_10tokens_completions(completions: list[str], **kwargs):
    return _compute_completion_sharp_target_token_length_reward(
        completions, target_tokens=10
    )


@register("sharp_20tokens_completions", RegistryType.REWARD_FUNCTION)
def _sharp_20tokens_completions(completions: list[str], **kwargs):
    return _compute_completion_sharp_target_token_length_reward(
        completions, target_tokens=20
    )



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import functools
import json
from typing import Any, NamedTuple

import pydantic
from typing_extensions import override

from oumi.builders import (
    build_processor,
    build_tokenizer,
    is_image_text_llm,
)
from oumi.core.configs import (
    GenerationParams,
    ModelParams,
    RemoteParams,
)
from oumi.core.configs.internal.supported_models import (
    find_internal_model_config_using_model_name,
)
from oumi.core.processors.base_processor import BaseProcessor
from oumi.core.types.conversation import Conversation, Message, Role, Type
from oumi.inference.remote_inference_engine import RemoteInferenceEngine
from oumi.utils.conversation_utils import (
    base64encode_content_item_image_bytes,
    load_image_bytes_to_content_item,
)


class _SamplingParams(NamedTuple):
    """It's a clone of `sglang.lang.ir.SglSamplingParams`.

    Only includes a subset of parameters supported in oumi.
    Unsupported params are left commented out for reference.
    """

    max_new_tokens: int = 128
    # min_new_tokens: int = 0
    stop: str | list[str] = ""
    stop_token_ids: list[int] | None = None
    temperature: float = 1.0
    top_p: float = 1.0
    # top_k: int = -1  # -1 means disable
    min_p: float = 0.0
    frequency_penalty: float = 0.0
    presence_penalty: float = 0.0
    # ignore_eos: bool = False
    # return_logprob: bool | None = None
    # logprob_start_len: int | None = None
    # top_logprobs_num: int | None = None
    # return_text_in_logprobs: bool | None = None
    json_schema: str | None = None

    # For constrained generation:
    # dtype: str | None = None
    regex: str | None = None


class SGLangInferenceEngine(RemoteInferenceEngine):
    """Engine for running SGLang inference."""

    def __init__(
        self,
        model_params: ModelParams,
        *,
        remote_params: RemoteParams | None = None,
        generation_params: GenerationParams | None = None,
    ):
        """Initializes the SGL inference Engine.

        Args:
            model_params: The model parameters to use for inference.
            remote_params: Remote server params.
            generation_params: The generation parameters to use for inference.
        """
        if remote_params is None:
            raise ValueError("remote_params is required")

        super().__init__(
            model_params=model_params,
            generation_params=generation_params,
            remote_params=remote_params,
        )

        self._tokenizer = build_tokenizer(self._model_params)
        self._processor: BaseProcessor | None = None
        self._supports_multiple_images: bool = False

        if is_image_text_llm(self._model_params):
            # Only enable Processor for vision language models for now.
            self._processor = build_processor(
                self._model_params.model_name,
                self._tokenizer,
                trust_remote_code=self._model_params.trust_remote_code,
                processor_kwargs=self._model_params.processor_kwargs,
            )
            internal_model_config = find_internal_model_config_using_model_name(
                self._model_params.model_name,
                trust_remote_code=self._model_params.trust_remote_code,
            )
            self._supports_multiple_images = (
                (internal_model_config is not None)
                and (internal_model_config.visual_config is not None)
                and internal_model_config.visual_config.supports_multiple_images
            )

        # TODO Launch a local SGLLang server if requested.

    def _create_sampling_params(
        self, generation_params: GenerationParams
    ) -> _SamplingParams:
        regex: str | None = None
        json_schema: str | None = None
        if generation_params.guided_decoding is not None:
            if generation_params.guided_decoding.regex is not None:
                regex = generation_params.guided_decoding.regex
            else:
                json_schema_value = None
                if generation_params.guided_decoding.json is not None:
                    json_schema_value = generation_params.guided_decoding.json
                elif (
                    generation_params.guided_decoding.choice is not None
                    and len(generation_params.guided_decoding.choice) > 0
                ):
                    json_schema_value = {
                        "enum": generation_params.guided_decoding.choice
                    }

                if isinstance(json_schema_value, str):
                    json_schema = json_schema_value
                elif isinstance(json_schema_value, dict):
                    json_schema = json.dumps(json_schema_value, ensure_ascii=False)
                elif isinstance(json_schema_value, pydantic.BaseModel) or (
                    isinstance(json_schema_value, type)
                    and issubclass(json_schema_value, pydantic.BaseModel)
                ):
                    json_schema = json.dumps(json_schema_value.model_json_schema())
                else:
                    raise ValueError(
                        "Unsupported type of generation_params.guided_decoding.json: "
                        f"{type(generation_params.guided_decoding.json)}"
                    )

        return _SamplingParams(
            max_new_tokens=generation_params.max_new_tokens,
            temperature=generation_params.temperature,
            top_p=generation_params.top_p,
            min_p=generation_params.min_p,
            frequency_penalty=generation_params.frequency_penalty,
            presence_penalty=generation_params.presence_penalty,
            stop=(generation_params.stop_strings or []),
            stop_token_ids=generation_params.stop_token_ids,
            regex=regex,
            json_schema=json_schema,
        )

    def _create_sampling_params_as_dict(
        self, generation_params: GenerationParams
    ) -> dict[str, Any]:
        return self._create_sampling_params(generation_params)._asdict()

    def _apply_chat_template_impl(self, conversation: Conversation) -> str:
        if self._processor is None:
            prompt = self._tokenizer.apply_chat_template(
                conversation.to_dict()["messages"],
                tokenize=False,
                add_generation_prompt=True,
            )

            if not isinstance(prompt, str):
                raise RuntimeError(
                    "`apply_chat_template` returned an object that is not a string. "
                    f"Actual type: {type(prompt)}"
                )
            return prompt

        return self._processor.apply_chat_template(
            conversation.messages,
            add_generation_prompt=True,
        )

    def _create_image_data_as_str_list(self, conversation: Conversation) -> list[str]:
        image_items = [
            item for m in conversation.messages for item in m.image_content_items
        ]
        num_images = len(image_items)
        if num_images <= 0:
            return []

        max_images = num_images if self._supports_multiple_images else 1

        if num_images > max_images:
            # If a conversation contains too many images, raise an error.
            # We can't silently discard extra images at this point
            # as many models verify that the actual number of images matches
            # the number of image tokens in text prompt.
            raise ValueError(
                conversation.append_id_to_string(
                    f"A conversation contains too many images ({num_images}). "
                    f"Max {max_images} image is allowed."
                )
            )

        result: list[str] = []
        for idx, image_item in enumerate(image_items):
            if image_item.type == Type.IMAGE_URL:
                # Preserve URL-s: leave them to SGLang server to download
                # to keep message payload size under control.
                # TODO Consider making this behaviour configurable.
                image_url = image_item.content
                if not image_url:
                    raise ValueError(
                        conversation.append_id_to_string(
                            f"Empty image URL in message: {image_item.type} "
                            f"in image item {idx + 1} of {num_images}!"
                        )
                    )
                result.append(image_url)
            else:
                image_item = load_image_bytes_to_content_item(image_item)
                if image_item.binary is None or len(image_item.binary) == 0:
                    raise ValueError(
                        conversation.append_id_to_string(
                            f"No image bytes in image item {idx + 1} of {num_images}!"
                        )
                    )
                result.append(base64encode_content_item_image_bytes(image_item))

        return result

    @override
    def _convert_conversation_to_api_input(
        self,
        conversation: Conversation,
        generation_params: GenerationParams,
        model_params: ModelParams,
    ) -> dict[str, Any]:
        """Converts a conversation to SGLang Native API input.

        See https://sgl-project.github.io/references/sampling_params.html for details.

        Args:
            conversation: The Oumi Conversation object to convert.
            generation_params: Parameters for text generation.
            model_params: Ignored.

        Returns:
            Dict[str, Any]: A dictionary containing the formatted input for the
            SGLang server native API, including the model, messages, generation params.
        """
        # Chat templates loaded by SGLang server are generally different from Oumi's
        # chat templates, hence, let's apply Oumi chat template here ourselves.
        prompt = self._apply_chat_template_impl(conversation)

        sampling_params_dict = self._create_sampling_params_as_dict(generation_params)
        body = {
            "text": prompt,
            "sampling_params": sampling_params_dict,
        }
        image_data: list[str] = self._create_image_data_as_str_list(conversation)
        if len(image_data) > 0:
            body["image_data"] = image_data if len(image_data) > 1 else image_data[0]
        return body

    @override
    def _convert_api_output_to_conversation(
        self, response: dict[str, Any], original_conversation: Conversation
    ) -> Conversation:
        """Converts an SGLang Native API response to a conversation."""
        new_message = Message(
            content=response["text"],
            role=Role.ASSISTANT,
        )
        return Conversation(
            messages=[*original_conversation.messages, new_message],
            metadata=original_conversation.metadata,
            conversation_id=original_conversation.conversation_id,
        )

    @override
    def _get_request_headers(self, remote_params: RemoteParams) -> dict[str, str]:
        return {
            "Content-Type": "application/json",
        }

    @override
    @functools.cache
    def get_supported_params(self) -> set[str]:
        """Returns a set of supported generation parameters for this engine."""
        return {
            "frequency_penalty",
            "guided_decoding",
            "max_new_tokens",
            "min_p",
            "presence_penalty",
            "stop_strings",
            "stop_token_ids",
            "temperature",
            "top_p",
        }



import logging
from unittest.mock import patch

import pytest
import torch

from oumi.builders.lr_schedules import (
    build_lr_scheduler,
)
from oumi.core.configs import SchedulerType, TrainingParams

#
# Fixtures
#


@pytest.fixture
def optimizer():
    params = [torch.nn.Parameter(torch.randn(2, 2), requires_grad=True)]
    return torch.optim.Adam(params=params, lr=0.001)


@pytest.fixture
def training_params():
    return TrainingParams(
        lr_scheduler_type=SchedulerType.LINEAR,
        warmup_steps=10,
        warmup_ratio=None,
        lr_scheduler_kwargs={},
        learning_rate=0.001,
    )


# Tests
#
@pytest.mark.parametrize(
    "scheduler_type",
    [
        SchedulerType.LINEAR,
        SchedulerType.COSINE,
        SchedulerType.CONSTANT,
        SchedulerType.COSINE_WITH_RESTARTS,
    ],
)
def test_build_schedulers(scheduler_type, optimizer, training_params):
    training_params.lr_scheduler_type = scheduler_type
    num_training_steps = 1000
    scheduler = build_lr_scheduler(optimizer, training_params, num_training_steps)
    assert isinstance(scheduler, torch.optim.lr_scheduler.LambdaLR)


def test_build_schedulers_with_unknown_type(optimizer, training_params):
    training_params.lr_scheduler_type = "unknown_type"
    with pytest.raises(ValueError, match="Unknown scheduler type"):
        build_lr_scheduler(optimizer, training_params)


@pytest.mark.parametrize(
    ("scheduler_type", "missing_ok"),
    [
        (SchedulerType.LINEAR, False),
        (SchedulerType.COSINE, False),
        (SchedulerType.CONSTANT, True),
        (SchedulerType.COSINE_WITH_RESTARTS, False),
    ],
)
def test_missing_num_training_steps_for_scheduler(
    scheduler_type, missing_ok, optimizer, training_params
):
    training_params.lr_scheduler_type = scheduler_type

    if missing_ok:
        scheduler = build_lr_scheduler(optimizer, training_params)
        assert isinstance(scheduler, torch.optim.lr_scheduler.LambdaLR)
    else:
        with pytest.raises(ValueError, match="num_training_steps must be provided"):
            build_lr_scheduler(optimizer, training_params)


@pytest.mark.parametrize(
    "scheduler_type",
    [
        SchedulerType.LINEAR,
        SchedulerType.COSINE,
        SchedulerType.CONSTANT,
        SchedulerType.COSINE_WITH_RESTARTS,
    ],
)
def test_warmup_ratio(scheduler_type, optimizer, training_params):
    num_training_steps = 1000
    training_params.warmup_steps = None
    training_params.warmup_ratio = 0.1
    training_params.lr_scheduler_type = scheduler_type
    scheduler = build_lr_scheduler(optimizer, training_params, num_training_steps)
    assert isinstance(scheduler, torch.optim.lr_scheduler.LambdaLR)


@pytest.mark.parametrize(
    "scheduler_type",
    [
        SchedulerType.LINEAR,
        SchedulerType.COSINE,
        SchedulerType.CONSTANT,
        SchedulerType.COSINE_WITH_RESTARTS,
    ],
)
def test_missing_num_training_steps_for_warmup_ratio(
    scheduler_type, optimizer, training_params
):
    training_params.warmup_steps = None
    training_params.warmup_ratio = 0.1
    training_params.lr_scheduler_type = scheduler_type
    with pytest.raises(ValueError, match="num_training_steps must be provided"):
        build_lr_scheduler(optimizer, training_params)


def test_both_warmup_steps_and_ratio_provided(optimizer, training_params):
    training_params.warmup_steps = 100
    training_params.warmup_ratio = 0.1
    with pytest.raises(
        ValueError, match="Only one of warmup_steps and warmup_ratio should be provided"
    ):
        build_lr_scheduler(optimizer, training_params, num_training_steps=1000)


def test_invalid_warmup_ratio(optimizer, training_params):
    training_params.warmup_steps = None
    training_params.warmup_ratio = 1.5
    with pytest.raises(ValueError, match=r"warmup_ratio must be in \[0, 1\]"):
        build_lr_scheduler(optimizer, training_params, num_training_steps=1000)


def test_no_warmup_provided(optimizer, training_params):
    training_params.warmup_steps = None
    training_params.warmup_ratio = None
    scheduler = build_lr_scheduler(optimizer, training_params, num_training_steps=1000)
    assert isinstance(scheduler, torch.optim.lr_scheduler.LambdaLR)


def test_linear_scheduler_params(optimizer, training_params):
    num_training_steps = 1000
    # Do epoch 0 first to initialize `initial_lr` in param_groups
    with patch(
        "oumi.builders.lr_schedules.get_linear_schedule_with_warmup"
    ) as mock_get_linear:
        build_lr_scheduler(optimizer, training_params, num_training_steps, 0)
        mock_get_linear.assert_called()
        mock_get_linear.assert_called_once_with(
            optimizer=optimizer,
            num_warmup_steps=10,
            num_training_steps=num_training_steps,
            last_epoch=-1,
        )

    current_epoch = 5
    with patch(
        "oumi.builders.lr_schedules.get_linear_schedule_with_warmup"
    ) as mock_get_linear:
        build_lr_scheduler(
            optimizer, training_params, num_training_steps, current_epoch
        )
        mock_get_linear.assert_called()
        mock_get_linear.assert_called_once_with(
            optimizer=optimizer,
            num_warmup_steps=10,
            num_training_steps=num_training_steps,
            last_epoch=4,
        )


def test_cosine_scheduler_params(optimizer, training_params):
    training_params.lr_scheduler_type = SchedulerType.COSINE
    num_training_steps = 1000

    # Do epoch 0 first to initialize `initial_lr` in param_groups
    with patch(
        "oumi.builders.lr_schedules.get_cosine_schedule_with_warmup"
    ) as mock_get_cosine:
        build_lr_scheduler(optimizer, training_params, num_training_steps, 0)
        mock_get_cosine.assert_called()
        mock_get_cosine.assert_called_once_with(
            optimizer=optimizer,
            num_warmup_steps=10,
            num_training_steps=num_training_steps,
            last_epoch=-1,
            num_cycles=0.5,
        )

    current_epoch = 5
    with patch(
        "oumi.builders.lr_schedules.get_cosine_schedule_with_warmup"
    ) as mock_get_cosine:
        build_lr_scheduler(
            optimizer, training_params, num_training_steps, current_epoch
        )
        mock_get_cosine.assert_called()
        mock_get_cosine.assert_called_once_with(
            optimizer=optimizer,
            num_warmup_steps=10,
            num_training_steps=num_training_steps,
            last_epoch=4,
            num_cycles=0.5,
        )


def test_cosine_with_restarts_scheduler_params(optimizer, training_params):
    training_params.lr_scheduler_type = SchedulerType.COSINE_WITH_RESTARTS
    num_training_steps = 1000
    training_params.lr_scheduler_kwargs = {"num_cycles": 3}
    # Do epoch 0 first to initialize `initial_lr` in param_groups
    with patch(
        "oumi.builders.lr_schedules.get_cosine_with_hard_restarts_schedule_with_warmup"
    ) as mock_get_cosine_restarts:
        build_lr_scheduler(optimizer, training_params, num_training_steps, 0)
        mock_get_cosine_restarts.assert_called()
        mock_get_cosine_restarts.assert_called_once_with(
            optimizer=optimizer,
            num_warmup_steps=10,
            num_training_steps=num_training_steps,
            last_epoch=-1,
            num_cycles=3,
        )

    current_epoch = 5
    with patch(
        "oumi.builders.lr_schedules.get_cosine_with_hard_restarts_schedule_with_warmup"
    ) as mock_get_cosine_restarts:
        build_lr_scheduler(
            optimizer, training_params, num_training_steps, current_epoch
        )
        mock_get_cosine_restarts.assert_called()
        mock_get_cosine_restarts.assert_called_once_with(
            optimizer=optimizer,
            num_warmup_steps=10,
            num_training_steps=num_training_steps,
            last_epoch=4,
            num_cycles=3,
        )


@pytest.mark.parametrize(
    "scheduler_type",
    [
        SchedulerType.LINEAR,
        SchedulerType.COSINE,
        SchedulerType.CONSTANT,
        SchedulerType.COSINE_WITH_RESTARTS,
    ],
)
def test_scheduler_specific_kwargs_warning(
    scheduler_type, optimizer, training_params, caplog
):
    training_params.lr_scheduler_kwargs = {"unknown_param": 42}
    training_params.lr_scheduler_type = scheduler_type

    # Enable propagation of log messages to the pytest logger for testing
    LOGGER = logging.getLogger("oumi")
    LOGGER.propagate = True

    with caplog.at_level("WARNING"):
        build_lr_scheduler(optimizer, training_params, num_training_steps=1000)
        assert "Unrecognized scheduler kwargs" in caplog.text


@patch("oumi.utils.logging.logger.info")
def test_warmup_ratio_logging(mock_logger_info, optimizer, training_params):
    training_params.warmup_steps = None
    training_params.warmup_ratio = 0.1
    num_training_steps = 1000
    build_lr_scheduler(optimizer, training_params, num_training_steps)
    mock_logger_info.assert_called_with(
        "Using warmup_steps=100 based on 0.1 warmup_ratio and 1000 max steps."
    )


@patch("oumi.utils.logging.logger.info")
def test_no_warmup_logging(mock_logger_info, optimizer, training_params):
    training_params.warmup_steps = None
    training_params.warmup_ratio = None
    build_lr_scheduler(optimizer, training_params, num_training_steps=1000)
    mock_logger_info.assert_called_with(
        "No warmup steps provided. Setting warmup_steps=0."
    )


def test_linear_scheduler_lr_values(optimizer, training_params):
    num_training_steps = 100
    scheduler = build_lr_scheduler(optimizer, training_params, num_training_steps)
    scheduler.step()

    # Check initial LR
    assert optimizer.param_groups[0]["lr"] == 0.0001  # 10% of 0.001 (warmup)

    # Move to end of warmup
    for _ in range(10 - 1):
        scheduler.step()
    assert optimizer.param_groups[0]["lr"] == 0.001  # Full LR

    # Move to end of training
    for _ in range(90):
        scheduler.step()
    assert pytest.approx(optimizer.param_groups[0]["lr"], 0.0001) == 0.0  # Final LR


def test_cosine_scheduler_lr_values(optimizer, training_params):
    training_params.lr_scheduler_type = SchedulerType.COSINE
    num_training_steps = 100
    scheduler = build_lr_scheduler(optimizer, training_params, num_training_steps)
    scheduler.step()

    # Check initial LR
    assert (
        pytest.approx(optimizer.param_groups[0]["lr"], 0.0001) == 0.0001
    )  # 10% of 0.001 (warmup)

    # Move to end of warmup
    for _ in range(10 - 1):
        scheduler.step()
    assert pytest.approx(optimizer.param_groups[0]["lr"], 0.0001) == 0.001  # Full LR

    # Move to end of training
    for _ in range(90):
        scheduler.step()
    assert pytest.approx(optimizer.param_groups[0]["lr"], 0.0001) == 0.0  # Final LR


def test_constant_scheduler_lr_values(optimizer, training_params):
    training_params.lr_scheduler_type = SchedulerType.CONSTANT
    scheduler = build_lr_scheduler(optimizer, training_params)
    scheduler.step()

    # Check initial LR
    assert (
        pytest.approx(optimizer.param_groups[0]["lr"], 0.0001) == 0.0001
    )  # 10% of 0.001 (warmup)

    # Move to end of warmup
    for _ in range(10 - 1):
        scheduler.step()
    assert pytest.approx(optimizer.param_groups[0]["lr"], 0.0001) == 0.001  # Full LR

    # Move further (should remain constant)
    for _ in range(90):
        scheduler.step()
    assert (
        pytest.approx(optimizer.param_groups[0]["lr"], 0.0001) == 0.001
    )  # Still full LR



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib.metadata
import importlib.util
import os
import platform

from rich.table import Table

import oumi.cli.cli_utils as cli_utils


def _get_package_version(package_name: str, version_fallback: str) -> str:
    """Gets the version of the specified package.

    Args:
        package_name: The name of the package.
        version_fallback: The fallback version string.

    Returns:
        str: The version of the package, or a fallback string if the package is not
            installed.
    """
    try:
        return importlib.metadata.version(package_name)
    except importlib.metadata.PackageNotFoundError:
        return version_fallback


def env():
    """Prints information about the current environment."""
    # Delayed imports
    from oumi.utils.torch_utils import format_cudnn_version
    # End imports

    version_fallback = "<not installed>"
    env_var_fallback = "<not set>"

    # All relevant environment vars.
    env_vars = sorted(
        [
            "ACCELERATE_DYNAMO_BACKEND",
            "ACCELERATE_DYNAMO_MODE",
            "ACCELERATE_DYNAMO_USE_FULLGRAPH",
            "ACCELERATE_DYNAMO_USE_DYNAMIC",
            "ACCELERATE_USE_FSDP",
            "CUDA_VISIBLE_DEVICES",
            "LOCAL_RANK",
            "LOCAL_WORLD_SIZE",
            "OUMI_EXTRA_DEPS_FILE",
            "OUMI_SLURM_CONNECTIONS",
            "OUMI_FORCE_EDITABLE_INSTALL",
            "OUMI_USE_SPOT_VM",
            "RANK",
            "WORLD_SIZE",
        ]
    )

    # All deps, excluding dev, docs, and gcp.
    core_packages = sorted(
        [
            "accelerate",
            "aiohttp",
            "bitsandbytes",
            "datasets",
            "diffusers",
            "einops",
            "jsonlines",
            "llama-cpp-python",
            "liger-kernel",
            "lm-eval",
            "numpy",
            "nvidia-ml-py",
            "omegaconf",
            "open_clip_torch",
            "pandas",
            "peft",
            "pexpect",
            "pillow",
            "pydantic",
            "responses",
            "sglang",
            "skypilot",
            "tensorboard",
            "timm",
            "torch",
            "torchdata",
            "torchvision",
            "tqdm",
            "transformers",
            "trl",
            "typer",
            "vllm",
            "wandb",
            "mlflow",
        ]
    )
    package_versions = {
        package: _get_package_version(package, version_fallback)
        for package in core_packages
    }
    env_values = {env_var: os.getenv(env_var, env_var_fallback) for env_var in env_vars}
    cli_utils.section_header("Oumi environment information:")
    env_table = Table(show_header=False, show_lines=False)
    env_table.add_row("Oumi version", _get_package_version("oumi", version_fallback))
    env_table.add_row("Python version", platform.python_version())
    env_table.add_row("Platform", platform.platform())
    cli_utils.CONSOLE.print(env_table)
    cli_utils.section_header("Installed dependencies:")
    deps_table = Table(show_header=True, show_lines=False)
    deps_table.add_column("PACKAGE", justify="left")
    deps_table.add_column("VERSION", justify="left")
    for package, version in package_versions.items():
        deps_table.add_row(package, version)
    cli_utils.CONSOLE.print(deps_table)

    if env_vars:
        cli_utils.section_header("Environment variables:")
        env_var_table = Table(show_header=True, show_lines=False)
        env_var_table.add_column("VARIABLE", justify="left")
        env_var_table.add_column("VALUE", justify="left")
        for var in env_vars:
            env_var_table.add_row(var, env_values[var])
        cli_utils.CONSOLE.print(env_var_table)

    if importlib.util.find_spec("torch") is not None:
        torch = importlib.import_module("torch")
        cli_utils.section_header("PyTorch information:")
        cuda_table = Table(show_header=False, show_lines=False)
        cuda_table.add_row("CUDA available", str(torch.cuda.is_available()))
        if torch.cuda.is_available():
            cuda_table.add_row("CUDA version", str(torch.version.cuda))
            cuda_table.add_row(
                "cuDNN version", format_cudnn_version(torch.backends.cudnn.version())
            )
            cuda_table.add_row("Number of GPUs", str(torch.cuda.device_count()))
            cuda_table.add_row("GPU type", torch.cuda.get_device_name())
            total_memory_gb = float(torch.cuda.mem_get_info()[1]) / float(
                1024 * 1024 * 1024
            )
            cuda_table.add_row("GPU memory", f"{total_memory_gb:.1f}GB")
        cli_utils.CONSOLE.print(cuda_table)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import tempfile
from pathlib import Path
from unittest.mock import patch

import pytest
import typer
from typer.testing import CliRunner

import oumi
from oumi.cli.cli_utils import CONTEXT_ALLOW_EXTRA_ARGS
from oumi.cli.synth import synth
from oumi.core.configs import (
    InferenceConfig,
    ModelParams,
    SynthesisConfig,
)
from oumi.core.configs.params.synthesis_params import GeneralSynthesisParams
from oumi.utils.logging import logger

runner = CliRunner()


def _create_synthesis_config() -> SynthesisConfig:
    return SynthesisConfig(
        strategy_params=GeneralSynthesisParams(),
        inference_config=InferenceConfig(
            model=ModelParams(
                model_name="MlpEncoder",
                trust_remote_code=True,
                tokenizer_name="gpt2",
            ),
        ),
        num_samples=5,
        output_path="test_output.jsonl",
    )


@pytest.fixture
def app():
    fake_app = typer.Typer()
    fake_app.command(context_settings=CONTEXT_ALLOW_EXTRA_ARGS)(synth)
    yield fake_app


@pytest.fixture
def mock_synthesize():
    with patch.object(oumi, "synthesize") as m_synthesize:
        yield m_synthesize


@pytest.fixture
def mock_fetch():
    with patch("oumi.cli.cli_utils.resolve_and_fetch_config") as m_fetch:
        yield m_fetch


@pytest.fixture
def mock_parse_extra_cli_args():
    with patch("oumi.cli.cli_utils.parse_extra_cli_args") as m_parse:
        m_parse.return_value = {}
        yield m_parse


@pytest.fixture
def mock_synthesis_config_from_yaml():
    with patch(
        "oumi.core.configs.synthesis_config.SynthesisConfig.from_yaml_and_arg_list"
    ) as m_config:
        yield m_config


@pytest.fixture
def sample_synthesis_results():
    return [
        {"prompt": "What is AI?", "response": "Artificial Intelligence..."},
        {"prompt": "What is ML?", "response": "Machine Learning..."},
        {"prompt": "What is DL?", "response": "Deep Learning..."},
    ]


def test_synth_runs(
    app,
    mock_synthesize,
    mock_parse_extra_cli_args,
    mock_synthesis_config_from_yaml,
    sample_synthesis_results,
):
    """Test that synth command runs successfully with basic configuration."""
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "synth.yaml")
        config: SynthesisConfig = _create_synthesis_config()
        config.to_yaml(yaml_path)

        mock_synthesis_config_from_yaml.return_value = config
        mock_synthesize.return_value = sample_synthesis_results

        result = runner.invoke(app, ["--config", yaml_path])

        assert result.exit_code == 0
        mock_parse_extra_cli_args.assert_called_once()
        mock_synthesis_config_from_yaml.assert_called_once_with(
            yaml_path, {}, logger=logger
        )
        mock_synthesize.assert_called_once_with(config)


def test_synth_with_overrides(
    app,
    mock_synthesize,
    mock_parse_extra_cli_args,
    mock_synthesis_config_from_yaml,
    sample_synthesis_results,
):
    """Test synth command with CLI argument overrides."""
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "synth.yaml")
        config: SynthesisConfig = _create_synthesis_config()
        config.to_yaml(yaml_path)

        expected_config = _create_synthesis_config()
        expected_config.num_samples = 10
        expected_config.output_path = "custom_output.jsonl"

        mock_synthesis_config_from_yaml.return_value = expected_config
        mock_synthesize.return_value = sample_synthesis_results

        result = runner.invoke(
            app,
            [
                "--config",
                yaml_path,
                "--num_samples",
                "10",
                "--output_path",
                "custom_output.jsonl",
            ],
        )

        assert result.exit_code == 0
        mock_synthesize.assert_called_once_with(expected_config)


def test_synth_with_oumi_prefix(
    app,
    mock_synthesize,
    mock_synthesis_config_from_yaml,
    mock_fetch,
    sample_synthesis_results,
):
    """Test synth command with oumi:// prefixed config path."""
    with tempfile.TemporaryDirectory() as temp_dir:
        output_dir = Path(temp_dir)
        config_path = "oumi://configs/examples/synthesis/basic.yaml"
        expected_path = output_dir / "configs/examples/synthesis/basic.yaml"

        config = _create_synthesis_config()
        expected_path.parent.mkdir(parents=True, exist_ok=True)
        config.to_yaml(expected_path)

        mock_fetch.return_value = expected_path
        mock_synthesis_config_from_yaml.return_value = config
        mock_synthesize.return_value = sample_synthesis_results

        with patch.dict("os.environ", {"OUMI_DIR": str(output_dir)}):
            result = runner.invoke(app, ["--config", config_path])

        assert result.exit_code == 0
        mock_fetch.assert_called_once_with(config_path)
        mock_synthesize.assert_called_once_with(config)


def test_synth_auto_generates_output_path(
    app,
    mock_synthesize,
    mock_synthesis_config_from_yaml,
    sample_synthesis_results,
):
    """Test that synth auto-generates output path when not specified."""
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "synth.yaml")
        config: SynthesisConfig = _create_synthesis_config()
        config.output_path = None  # No output path specified
        config.to_yaml(yaml_path)

        mock_synthesis_config_from_yaml.return_value = config
        mock_synthesize.return_value = sample_synthesis_results

        with (
            patch("oumi.cli.synth.Path.cwd") as mock_cwd,
            patch("oumi.cli.synth.datetime") as mock_datetime,
        ):
            mock_cwd.return_value = Path(output_temp_dir)
            mock_datetime.now.return_value.strftime.return_value = "20250124_120000"

            result = runner.invoke(app, ["--config", yaml_path])

            assert result.exit_code == 0
            # Verify that output_path was set on the config
            args, _ = mock_synthesize.call_args
            actual_config = args[0]
            assert actual_config.output_path.endswith(
                "oumi_synth_results_20250124_120000.jsonl"
            )


def test_synth_logging_levels(
    app,
    mock_synthesize,
    mock_synthesis_config_from_yaml,
    sample_synthesis_results,
):
    """Test synth command with different logging levels."""
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "synth.yaml")
        config: SynthesisConfig = _create_synthesis_config()
        config.to_yaml(yaml_path)

        mock_synthesis_config_from_yaml.return_value = config
        mock_synthesize.return_value = sample_synthesis_results

        result = runner.invoke(app, ["--config", yaml_path, "--log-level", "WARNING"])
        assert result.exit_code == 0
        assert logger.level == logging.WARNING

        result = runner.invoke(app, ["--config", yaml_path, "-log", "CRITICAL"])
        assert result.exit_code == 0
        assert logger.level == logging.CRITICAL


def test_synth_fails_no_config(app):
    """Test that synth command fails when no config is provided."""
    result = runner.invoke(app, [])
    assert result.exit_code == 2


def test_synth_with_empty_results(
    app,
    mock_synthesize,
    mock_synthesis_config_from_yaml,
):
    """Test synth command handles empty synthesis results."""
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "synth.yaml")
        config: SynthesisConfig = _create_synthesis_config()
        config.to_yaml(yaml_path)

        mock_synthesis_config_from_yaml.return_value = config
        mock_synthesize.return_value = []

        result = runner.invoke(app, ["--config", yaml_path])

        assert "No results found" in result.stdout
        assert result.exit_code == 0
        mock_synthesize.assert_called_once_with(config)


def test_synth_with_large_results(
    app,
    mock_synthesize,
    mock_synthesis_config_from_yaml,
):
    """Test synth command with more than 5 results (shows truncation message)."""
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "synth.yaml")
        config: SynthesisConfig = _create_synthesis_config()
        config.to_yaml(yaml_path)

        # Create 7 sample results
        large_results = [
            {"prompt": f"Question {i}?", "response": f"Answer {i}"} for i in range(7)
        ]

        mock_synthesis_config_from_yaml.return_value = config
        mock_synthesize.return_value = large_results

        result = runner.invoke(app, ["--config", yaml_path])

        assert result.exit_code == 0
        mock_synthesize.assert_called_once_with(config)
        assert "and 6 more samples" in result.stdout


def test_synth_config_validation_error(
    app,
    mock_synthesis_config_from_yaml,
):
    """Test synth command handles config validation errors."""
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "synth.yaml")

        # Mock config validation to raise an error
        mock_synthesis_config_from_yaml.side_effect = ValueError("Invalid config")

        result = runner.invoke(app, ["--config", yaml_path])

        # Should fail due to config validation error
        assert result.exit_code != 0


def test_synth_synthesis_error(
    app,
    mock_synthesize,
    mock_synthesis_config_from_yaml,
):
    """Test synth command handles synthesis errors."""
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "synth.yaml")
        config: SynthesisConfig = _create_synthesis_config()
        config.to_yaml(yaml_path)

        mock_synthesis_config_from_yaml.return_value = config
        mock_synthesize.side_effect = RuntimeError("Synthesis failed")

        result = runner.invoke(app, ["--config", yaml_path])

        # Should fail due to synthesis error
        assert result.exit_code != 0
        mock_synthesize.assert_called_once_with(config)


def test_synth_output_path_collision_handling(
    app,
    mock_synthesize,
    mock_synthesis_config_from_yaml,
    sample_synthesis_results,
):
    """Test that synth handles output path collisions by incrementing filename."""
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "synth.yaml")
        config: SynthesisConfig = _create_synthesis_config()
        config.output_path = None  # No output path specified
        config.to_yaml(yaml_path)

        mock_synthesis_config_from_yaml.return_value = config
        mock_synthesize.return_value = sample_synthesis_results

        # Create a file that would collide with the auto-generated name
        with (
            patch("oumi.cli.synth.Path.cwd") as mock_cwd,
            patch("oumi.cli.synth.datetime") as mock_datetime,
        ):
            mock_cwd.return_value = Path(output_temp_dir)
            mock_datetime.now.return_value.strftime.return_value = "20250124_120000"

            # Create the file that would be auto-generated
            collision_file = (
                Path(output_temp_dir) / "oumi_synth_results_20250124_120000.jsonl"
            )
            collision_file.touch()

            result = runner.invoke(app, ["--config", yaml_path])

            assert result.exit_code == 0
            # Verify that output_path was incremented to avoid collision
            args, _ = mock_synthesize.call_args
            actual_config = args[0]
            assert actual_config.output_path.endswith(
                "oumi_synth_results_20250124_120000_1.jsonl"
            )


def test_synth_config_finalization(
    app,
    mock_synthesize,
    mock_synthesis_config_from_yaml,
    sample_synthesis_results,
):
    """Test that synth calls config finalization and validation."""
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "synth.yaml")
        config: SynthesisConfig = _create_synthesis_config()
        config.to_yaml(yaml_path)

        mock_synthesis_config_from_yaml.return_value = config
        mock_synthesize.return_value = sample_synthesis_results

        with patch.object(config, "finalize_and_validate") as mock_finalize:
            result = runner.invoke(app, ["--config", yaml_path])

            assert result.exit_code == 0
            mock_finalize.assert_called_once()


def test_synth_short_flag_syntax(
    app,
    mock_synthesize,
    mock_synthesis_config_from_yaml,
    sample_synthesis_results,
):
    """Test synth command with short flag syntax."""
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "synth.yaml")
        config: SynthesisConfig = _create_synthesis_config()
        config.to_yaml(yaml_path)

        mock_synthesis_config_from_yaml.return_value = config
        mock_synthesize.return_value = sample_synthesis_results

        result = runner.invoke(app, ["-c", yaml_path])

        assert result.exit_code == 0
        mock_synthesize.assert_called_once_with(config)



import pytest
import torch

from oumi.utils.hf_utils import find_hf_token


def requires_gpus(count: int = 1, min_gb: float = 0.0) -> pytest.MarkDecorator:
    """Decorator to skip a test if the required number of GPUs is not available.

    Args:
        count (int): The number of GPUs required for the test. Defaults to 1.
        min_gb: Min required GPU VRAM in GB-s. Has no effect if zero or negative.

    Returns:
        pytest.MarkDecorator: A decorator that skips the test if the required
            number of GPUs is not available.
    """

    if not torch.cuda.is_available():
        return pytest.mark.skip(reason="CUDA is not available")

    gpu_count = torch.cuda.device_count()

    error_message = ""
    if gpu_count < count:
        error_message = (
            f"Not enough GPUs to run the test: requires '{count}',"
            f" got '{torch.cuda.device_count()}'"
        )
    elif min_gb > 0.0:
        eps = 1e-2  # relative tolerance
        for device_idx in range(gpu_count):
            _, total_memory = torch.cuda.mem_get_info(device_idx)
            total_memory_gb = float(total_memory) / float(1024 * 1024 * 1024)
            if total_memory_gb < min_gb * (1 - eps):
                device_name = torch.cuda.get_device_name(device_idx)
                error_message = (
                    "Not enough GPU memory to run the test: "
                    f"requires {min_gb:.3f}GB, got {total_memory_gb:.3f}GB. "
                    f"GPU: {device_name}"
                ) + (f" ({device_idx + 1} of {gpu_count})" if gpu_count > 1 else "")

    return pytest.mark.skipif(len(error_message) > 0, reason=error_message)


def requires_cuda_initialized() -> pytest.MarkDecorator:
    if not torch.cuda.is_available():
        return pytest.mark.skip(reason="CUDA is not available")

    if not torch.cuda.is_initialized():
        torch.cuda.init()

    return pytest.mark.skipif(
        not torch.cuda.is_initialized(), reason="CUDA is not initialized"
    )


def requires_cuda_not_available() -> pytest.MarkDecorator:
    return pytest.mark.skipif(torch.cuda.is_available(), reason="CUDA is available")


def requires_hf_token() -> pytest.MarkDecorator:
    return pytest.mark.skipif(not find_hf_token(), reason="HF token is not available")


def requires_pdf_support():
    from importlib.util import find_spec

    return pytest.mark.skipif(
        not find_spec("pdf2image"), reason="PDF support is not available"
    )



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oumi.core.datasets import BasePretrainingDataset
from oumi.core.registry import register_dataset


@register_dataset("togethercomputer/RedPajama-Data-1T")
class RedPajamaDataV1Dataset(BasePretrainingDataset):
    """RedPajama is a clean-room, fully open-source implementation of the LLaMa dataset.

    This dataset contains approximately 1.2 trillion tokens from various sources:
    Commoncrawl (878B), C4 (175B), GitHub (59B), ArXiv (28B), Wikipedia (24B),
    and StackExchange (20B) :footcite:`2023_redpajama`.

    The dataset is primarily in English, though the Wikipedia slice contains
    multiple languages.

    Dataset Structure:
        .. code-block:: python

            {
                "text": str,
                "meta": {
                    "url": str,
                    "timestamp": str,
                    "source": str,
                    "language": str,
                    ...
                },
                "red_pajama_subset": str
            }

    Subsets:
        - common_crawl
        - c4
        - github
        - arxiv
        - wikipedia
        - stackexchange

    See Also:
        - For more information on dataset creation and source data, please refer
          to the RedPajama GitHub repository:
          https://github.com/togethercomputer/RedPajama-Data
        - Hugging Face dataset page:
          https://huggingface.co/datasets/togethercomputer/RedPajama-Data-1T

    Note:
        The 'book' config is defunct and no longer accessible due to reported
        copyright infringement for the Book3 dataset contained in this config.

    Note:
        Please refer to the licenses of the data subsets you use. Links to the
        respective licenses can be found in the README.

    Citations:
        .. footbibliography::
    """

    default_dataset = "togethercomputer/RedPajama-Data-1T"



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import Mock, patch

import pytest

from oumi.core.configs.inference_config import InferenceConfig
from oumi.core.configs.inference_engine_type import InferenceEngineType
from oumi.core.configs.params.model_params import ModelParams
from oumi.core.configs.params.remote_params import RemoteParams
from oumi.core.configs.params.synthesis_params import (
    GeneralSynthesisParams,
    GeneratedAttribute,
    GeneratedAttributePostprocessingParams,
    SampledAttribute,
    SampledAttributeValue,
    TextMessage,
)
from oumi.core.synthesis.attribute_synthesizer import AttributeSynthesizer
from oumi.core.types.conversation import Conversation, Message, Role


@pytest.fixture
def mock_inference_config():
    """Create a mock inference config."""
    mock = Mock(spec=InferenceConfig)
    mock.engine = InferenceEngineType.NATIVE
    mock.model = Mock(spec=ModelParams)
    mock.remote_params = Mock(spec=RemoteParams)
    return mock


@pytest.fixture
def mock_permutable_attributes():
    """Create mock permutable attributes for testing."""
    return [
        SampledAttribute(
            id="style",
            name="Writing Style",
            description="The style of writing to use",
            possible_values=[
                SampledAttributeValue(
                    id="formal",
                    name="Formal",
                    description="A formal writing style",
                ),
                SampledAttributeValue(
                    id="casual",
                    name="Casual",
                    description="A casual writing style",
                ),
            ],
        ),
        SampledAttribute(
            id="topic",
            name="Topic",
            description="The topic to write about",
            possible_values=[
                SampledAttributeValue(
                    id="tech",
                    name="Technology",
                    description="Technology topics",
                ),
                SampledAttributeValue(
                    id="science",
                    name="Science",
                    description="Science topics",
                ),
            ],
        ),
    ]


@pytest.fixture
def mock_general_synthesis_params(mock_permutable_attributes):
    """Create mock GeneralSynthesisParams for testing."""
    return GeneralSynthesisParams(
        sampled_attributes=mock_permutable_attributes,
    )


@pytest.fixture
def mock_generated_attribute():
    """Create mock GeneratedAttribute for testing."""
    return GeneratedAttribute(
        id="generated_content",
        instruction_messages=[
            TextMessage(
                role=Role.SYSTEM,
                content="You are a helpful assistant.",
            ),
            TextMessage(
                role=Role.USER,
                content="Write a {style} paragraph about {topic}.",
            ),
        ],
    )


@pytest.fixture
def mock_samples():
    """Create mock samples for testing."""
    return [
        {"style": "formal", "topic": "tech"},
        {"style": "casual", "topic": "science"},
        {"non_permutable": "some_value"},
    ]


@patch("oumi.core.synthesis.attribute_synthesizer.build_inference_engine")
def test_synthesize_returns_dict_list(
    mock_build_inference_engine,
    mock_general_synthesis_params,
    mock_generated_attribute,
    mock_inference_config,
):
    """Test that synthesize returns list of dictionaries."""
    mock_inference_engine = Mock()
    mock_build_inference_engine.return_value = mock_inference_engine

    # Mock the inference engine's infer method to return conversations with responses
    mock_inference_engine.infer.return_value = [
        Conversation(
            messages=[
                Message(role=Role.USER, content="Test query"),
                Message(role=Role.ASSISTANT, content="Test response 1"),
            ]
        ),
        Conversation(
            messages=[
                Message(role=Role.USER, content="Test query"),
                Message(role=Role.ASSISTANT, content="Test response 2"),
            ]
        ),
    ]

    synthesizer = AttributeSynthesizer(
        mock_general_synthesis_params,
        mock_inference_config,
    )
    # Use samples that have all required fields
    samples = [
        {"style": "formal", "topic": "tech"},
        {"style": "casual", "topic": "science"},
    ]
    result = synthesizer.synthesize(samples, mock_generated_attribute)

    assert isinstance(result, list)
    assert len(result) == len(samples)
    for item in result:
        assert isinstance(item, dict)
        assert "generated_content" in item


@patch("oumi.core.synthesis.attribute_synthesizer.build_inference_engine")
def test_format_instructions_with_permutable_attributes(
    mock_formatter_class,
    mock_general_synthesis_params,
    mock_generated_attribute,
    mock_inference_config,
):
    """Test formatting instructions with permutable attributes."""
    # Mock the formatter instance
    mock_formatter = Mock()
    mock_formatter.format.side_effect = [
        "You are a helpful assistant.",
        "Write a Formal paragraph about Technology.",
    ]
    mock_formatter_class.return_value = mock_formatter

    synthesizer = AttributeSynthesizer(
        mock_general_synthesis_params,
        mock_inference_config,
    )
    sample = {"style": "formal", "topic": "tech"}

    result = synthesizer._format_instructions(
        sample,
        mock_generated_attribute.instruction_messages,
    )

    assert isinstance(result, Conversation)
    assert len(result.messages) == 2

    # Check that the formatting worked correctly
    assert result.messages[0].role == Role.SYSTEM
    assert result.messages[0].content == "You are a helpful assistant."
    assert result.messages[1].role == Role.USER
    assert result.messages[1].content == "Write a Formal paragraph about Technology."


@patch("oumi.core.synthesis.attribute_synthesizer.build_inference_engine")
def test_synthesize_with_multiple_samples(
    mock_build_inference_engine,
    mock_general_synthesis_params,
    mock_generated_attribute,
    mock_inference_config,
):
    """Test synthesize with multiple samples."""
    mock_inference_engine = Mock()
    mock_build_inference_engine.return_value = mock_inference_engine

    # Mock the inference engine's infer method to return conversations with responses
    mock_inference_engine.infer.return_value = [
        Conversation(
            messages=[
                Message(role=Role.USER, content="Test query"),
                Message(role=Role.ASSISTANT, content="Test response 1"),
            ]
        ),
        Conversation(
            messages=[
                Message(role=Role.USER, content="Test query"),
                Message(role=Role.ASSISTANT, content="Test response 2"),
            ]
        ),
    ]

    synthesizer = AttributeSynthesizer(
        mock_general_synthesis_params, mock_inference_config
    )
    samples = [
        {"style": "formal", "topic": "tech"},
        {"style": "casual", "topic": "science"},
    ]

    result = synthesizer.synthesize(samples, mock_generated_attribute)

    assert len(result) == 2
    for item in result:
        assert isinstance(item, dict)
        assert "generated_content" in item


@patch("oumi.core.synthesis.attribute_synthesizer.build_inference_engine")
def test_synthesize_with_empty_samples(
    mock_build_inference_engine,
    mock_general_synthesis_params,
    mock_generated_attribute,
    mock_inference_config,
):
    """Test synthesize with empty samples list."""
    mock_inference_engine = Mock()
    mock_build_inference_engine.return_value = mock_inference_engine

    # Mock the inference engine's infer method to return empty list
    mock_inference_engine.infer.return_value = []

    synthesizer = AttributeSynthesizer(
        mock_general_synthesis_params, mock_inference_config
    )
    samples = []

    result = synthesizer.synthesize(samples, mock_generated_attribute)

    assert result == []


@patch("oumi.core.synthesis.attribute_synthesizer.build_inference_engine")
def test_postprocess_sample(mock_build_inference_engine):
    """Test postprocessing a sample."""
    mock_build_inference_engine.return_value = Mock()

    synthesizer = AttributeSynthesizer(GeneralSynthesisParams(), Mock())

    response = "Response: Here is the formal text [END]"
    postprocessing_params = GeneratedAttributePostprocessingParams(
        id="processed_content",
        cut_prefix="Response: ",
        cut_suffix=" [END]",
        strip_whitespace=True,
        added_prefix="New: ",
        added_suffix=" (done)",
    )

    result = synthesizer._postprocess_sample(response, postprocessing_params)

    assert result == "New: Here is the formal text (done)"


@patch("oumi.core.synthesis.attribute_synthesizer.build_inference_engine")
def test_postprocess_sample_with_regex(mock_build_inference_engine):
    """Test postprocessing a sample with regex."""
    mock_build_inference_engine.return_value = Mock()

    synthesizer = AttributeSynthesizer(GeneralSynthesisParams(), Mock())

    response = "The answer is 42 and that's final."
    postprocessing_params = GeneratedAttributePostprocessingParams(
        id="processed_content",
        regex=r"\d+",
        added_prefix="Number: ",
    )

    result = synthesizer._postprocess_sample(response, postprocessing_params)

    assert result == "Number: 42"


@patch("oumi.core.synthesis.attribute_synthesizer.build_inference_engine")
def test_postprocess_sample_with_no_regex_match(mock_build_inference_engine):
    """Test postprocessing a sample when regex doesn't match."""
    mock_build_inference_engine.return_value = Mock()
    synthesizer = AttributeSynthesizer(GeneralSynthesisParams(), Mock())

    response = "No numbers here!"
    postprocessing_params = GeneratedAttributePostprocessingParams(
        id="processed_content",
        regex=r"\d+",
        added_prefix="Number: ",
    )

    result = synthesizer._postprocess_sample(response, postprocessing_params)

    assert result == "Number: No numbers here!"



import contextlib
from unittest import mock
from unittest.mock import patch

import pytest

from oumi.builders.inference_engines import ENGINE_MAP, build_inference_engine
from oumi.core.configs import (
    GenerationParams,
    InferenceConfig,
    InferenceEngineType,
    ModelParams,
    RemoteParams,
)
from oumi.inference import (
    AnthropicInferenceEngine,
    BedrockInferenceEngine,
    DeepSeekInferenceEngine,
    GoogleGeminiInferenceEngine,
    GoogleVertexInferenceEngine,
    LlamaCppInferenceEngine,
    NativeTextInferenceEngine,
    OpenAIInferenceEngine,
    ParasailInferenceEngine,
    RemoteInferenceEngine,
    RemoteVLLMInferenceEngine,
    SambanovaInferenceEngine,
    SGLangInferenceEngine,
    TogetherInferenceEngine,
    VLLMInferenceEngine,
)

# Check if optional dependencies are available
try:
    import vllm  # noqa: F401 # pyright: ignore[reportMissingImports]

    vllm_import_failed = False
except ImportError:
    vllm_import_failed = True

try:
    import llama_cpp  # noqa: F401 # pyright: ignore[reportMissingImports]

    llama_cpp_import_failed = False
except ImportError:
    llama_cpp_import_failed = True

try:
    import boto3  # noqa: F401 # pyright: ignore[reportMissingImports]

    boto3_import_failed = False
except ImportError:
    boto3_import_failed = True

# Group engines by whether they require remote params
LOCAL_ENGINES = [
    NativeTextInferenceEngine,
    VLLMInferenceEngine,
    LlamaCppInferenceEngine,
]

REMOTE_ENGINES = [
    RemoteInferenceEngine,
    RemoteVLLMInferenceEngine,
    SGLangInferenceEngine,
]

REMOTE_API_ENGINES = [
    AnthropicInferenceEngine,
    BedrockInferenceEngine,
    DeepSeekInferenceEngine,
    GoogleGeminiInferenceEngine,
    GoogleVertexInferenceEngine,
    OpenAIInferenceEngine,
    ParasailInferenceEngine,
    SambanovaInferenceEngine,
    TogetherInferenceEngine,
]


def _should_skip_engine(engine_class) -> bool:
    return (
        (engine_class == VLLMInferenceEngine and vllm_import_failed)
        or (engine_class == LlamaCppInferenceEngine and llama_cpp_import_failed)
        or (engine_class == BedrockInferenceEngine and boto3_import_failed)
    )


def _mock_engine(engine_class):
    """Mock the engine to avoid loading non-existent models."""

    mock_tokenizer = mock.MagicMock()
    mock_tokenizer.pad_token_id = 0
    mock_tokenizer.eos_token_id = 0
    mock_tokenizer.eos_token = "<eos>"
    mock_model = mock.MagicMock()
    mock_model.generate = mock.MagicMock()  # Add generate attribute

    if engine_class == VLLMInferenceEngine:
        mock_llm = mock.MagicMock()
        mock_ctx = patch.multiple(
            "oumi.inference.vllm_inference_engine",
            vllm=mock.MagicMock(LLM=mock.MagicMock(return_value=mock_llm)),
            build_tokenizer=mock.MagicMock(return_value=mock_tokenizer),
        )
    elif engine_class == LlamaCppInferenceEngine:
        mock_ctx = patch("llama_cpp.Llama.from_pretrained")
    elif engine_class == BedrockInferenceEngine:
        mock_ctx = patch("boto3.client")
    elif engine_class == SGLangInferenceEngine:
        mock_ctx = patch.multiple(
            "oumi.inference.sglang_inference_engine",
            build_tokenizer=mock.MagicMock(return_value=mock_tokenizer),
            build_processor=mock.MagicMock(return_value=None),
            is_image_text_llm=mock.MagicMock(return_value=False),
        )
    elif engine_class == NativeTextInferenceEngine:
        mock_ctx = patch.multiple(
            "oumi.inference.native_text_inference_engine",
            build_model=mock.MagicMock(return_value=mock_model),
            build_tokenizer=mock.MagicMock(return_value=mock_tokenizer),
            build_processor=mock.MagicMock(return_value=None),
            is_image_text_llm=mock.MagicMock(return_value=False),
        )
    elif issubclass(engine_class, RemoteInferenceEngine):
        mock_ctx = patch("aiohttp.ClientSession")
    else:
        mock_ctx = contextlib.nullcontext()

    return mock_ctx


@pytest.mark.parametrize("engine_class", LOCAL_ENGINES)
def test_local_engine_init_with_model_params(engine_class):
    """Test that local engines can be initialized with just model params."""
    if _should_skip_engine(engine_class):
        pytest.skip(
            f"Skipping {engine_class} because it is not supported on this platform"
        )
    model_params = ModelParams(model_name="test-model")
    mock_engine_class = _mock_engine(engine_class)
    with mock_engine_class:
        engine = engine_class(model_params=model_params)
    assert engine._model_params.model_name == "test-model"


@pytest.mark.parametrize("engine_class", REMOTE_ENGINES + REMOTE_API_ENGINES)
def test_remote_engine_init_with_model_and_remote_params(engine_class):
    """Test that remote engines can be initialized with both model and remote params."""
    if _should_skip_engine(engine_class):
        pytest.skip(
            f"Skipping {engine_class} because it is not supported on this platform"
        )
    model_params = ModelParams(model_name="test-model")
    remote_params = RemoteParams(api_url="http://test.com", api_key="test-key")
    mock_engine_class = _mock_engine(engine_class)
    with mock_engine_class:
        engine = engine_class(model_params=model_params, remote_params=remote_params)
    assert engine._model_params.model_name == "test-model"
    assert engine._remote_params.api_url == "http://test.com"
    assert engine._remote_params.api_key == "test-key"


@pytest.mark.parametrize(
    "engine_class", LOCAL_ENGINES + REMOTE_ENGINES + REMOTE_API_ENGINES
)
def test_engine_init_missing_model_params_fails(engine_class):
    """Test that all engines fail when initialized without model params."""
    remote_params = RemoteParams(api_url="http://test.com", api_key="test-key")
    with pytest.raises(TypeError):
        if engine_class in LOCAL_ENGINES:
            engine_class()  # Should fail - missing model_params
        else:
            engine_class(
                remote_params=remote_params
            )  # Should fail - missing model_params


@pytest.mark.parametrize(
    "engine_class", LOCAL_ENGINES + REMOTE_ENGINES + REMOTE_API_ENGINES
)
def test_engine_init_with_invalid_params_fails(engine_class):
    """Test that all engines fail when initialized with invalid params."""
    with pytest.raises(TypeError):
        engine_class(invalid_param="test")  # Should fail - invalid param


@pytest.mark.parametrize("engine_class", LOCAL_ENGINES)
def test_local_engine_config_overrides_constructor_params(engine_class):
    """Test that InferenceConfig params override constructor params."""
    if _should_skip_engine(engine_class):
        pytest.skip(
            f"Skipping {engine_class} because it is not supported on this platform"
        )
    # Initialize with one set of params
    init_model_params = ModelParams(
        model_name="init-model",
        model_max_length=128,
        torch_dtype_str="float32",
    )
    mock_engine_class = _mock_engine(engine_class)
    with mock_engine_class:
        engine = engine_class(model_params=init_model_params)
    assert engine._model_params.model_name == "init-model"

    # Create config with different params
    config_model_params = ModelParams(
        model_name="config-model",
        model_max_length=256,
        torch_dtype_str="float16",
    )
    config = InferenceConfig(
        model=config_model_params,
        generation=GenerationParams(max_new_tokens=100),
    )

    # Mock _infer to avoid actual inference
    with patch.object(engine, "_infer") as mock_infer:
        engine.infer([], config)

        # Verify the config params were used
        call_args = mock_infer.call_args[0]
        assert len(call_args) >= 2  # Should have at least input and config args
        passed_config = call_args[1]
        assert passed_config.model.model_name == "config-model"
        assert passed_config.model.model_max_length == 256
        assert passed_config.model.torch_dtype_str == "float16"


@pytest.mark.parametrize("engine_class", REMOTE_ENGINES + REMOTE_API_ENGINES)
def test_remote_engine_config_overrides_constructor_params(engine_class):
    """Test that InferenceConfig params override constructor params."""
    if _should_skip_engine(engine_class):
        pytest.skip(
            f"Skipping {engine_class} because it is not supported on this platform"
        )
    # Initialize with one set of params
    init_model_params = ModelParams(
        model_name="init-model",
        model_max_length=128,
    )
    init_remote_params = RemoteParams(
        api_url="http://init.com",
        api_key="init-key",
        num_workers=1,
        max_retries=3,
    )
    mock_engine_class = _mock_engine(engine_class)
    with mock_engine_class:
        engine = engine_class(
            model_params=init_model_params,
            remote_params=init_remote_params,
        )
    assert engine._model_params.model_name == "init-model"
    assert engine._remote_params.api_url == "http://init.com"

    # Create config with different params
    config_model_params = ModelParams(
        model_name="config-model",
        model_max_length=256,
    )
    config_remote_params = RemoteParams(
        api_url="http://config.com",
        api_key="config-key",
        num_workers=2,
        max_retries=5,
    )
    config = InferenceConfig(
        model=config_model_params,
        remote_params=config_remote_params,
        generation=GenerationParams(max_new_tokens=100),
    )

    # Mock _infer to avoid actual inference
    with patch.object(engine, "_infer") as mock_infer:
        engine.infer([], config)

        # Verify the config params were used
        call_args = mock_infer.call_args[0]
        assert len(call_args) >= 2  # Should have at least input and config args
        passed_config = call_args[1]
        assert passed_config.model.model_name == "config-model"
        assert passed_config.model.model_max_length == 256
        assert passed_config.remote_params.api_url == "http://config.com"
        assert passed_config.remote_params.api_key == "config-key"
        assert passed_config.remote_params.num_workers == 2
        assert passed_config.remote_params.max_retries == 5


@pytest.mark.parametrize(
    "engine_class", LOCAL_ENGINES + REMOTE_ENGINES + REMOTE_API_ENGINES
)
def test_engine_config_partial_override(engine_class):
    """Test that InferenceConfig partially overrides constructor params."""
    if _should_skip_engine(engine_class):
        pytest.skip(
            f"Skipping {engine_class} because it is not supported on this platform"
        )
    # Initialize with full params
    init_model_params = ModelParams(
        model_name="init-model",
        model_max_length=128,
        torch_dtype_str="float32",
    )
    init_remote_params = (
        RemoteParams(
            api_url="http://init.com",
            api_key="init-key",
            num_workers=1,
            max_retries=3,
        )
        if engine_class not in LOCAL_ENGINES
        else None
    )

    mock_engine_class = _mock_engine(engine_class)
    with mock_engine_class:
        if issubclass(engine_class, RemoteInferenceEngine):
            engine = engine_class(
                model_params=init_model_params,
                remote_params=init_remote_params,
            )
        else:
            engine = engine_class(model_params=init_model_params)

    # Create config with only some params changed
    config_model_params = ModelParams(
        model_name="config-model",  # Changed
        model_max_length=128,  # Same as init
        torch_dtype_str="float32",  # Same as init
    )
    config_remote_params = (
        RemoteParams(
            api_url="http://config.com",  # Changed
            api_key="init-key",  # Same as init
            num_workers=1,  # Same as init
            max_retries=3,  # Same as init
        )
        if engine_class not in LOCAL_ENGINES
        else None
    )
    config = InferenceConfig(
        model=config_model_params,
        remote_params=config_remote_params,
        generation=GenerationParams(max_new_tokens=100),
    )

    # Mock _infer to avoid actual inference
    with patch.object(engine, "_infer") as mock_infer:
        engine.infer([], config)

        # Verify only changed params were overridden
        call_args = mock_infer.call_args[0]
        passed_config = call_args[1]
        assert passed_config.model.model_name == "config-model"  # Changed
        assert passed_config.model.model_max_length == 128  # Same
        assert passed_config.model.torch_dtype_str == "float32"  # Same

        if engine_class not in LOCAL_ENGINES:
            assert passed_config.remote_params.api_url == "http://config.com"  # Changed
            assert passed_config.remote_params.api_key == "init-key"  # Same
            assert passed_config.remote_params.num_workers == 1  # Same
            assert passed_config.remote_params.max_retries == 3  # Same


def test_all_inference_engine_types_in_engine_map():
    """Test that all InferenceEngineType values are present in ENGINE_MAP."""
    for engine_type in InferenceEngineType:
        assert engine_type in ENGINE_MAP, (
            f"Missing engine type {engine_type} in ENGINE_MAP"
        )


def test_build_all_inference_engines():
    """Test that all inference engines can be built using the builder."""
    model_params = ModelParams(model_name="test-model")
    remote_params = RemoteParams(api_url="http://test.com", api_key="test-key")

    for engine_type in InferenceEngineType:
        engine_class = ENGINE_MAP[engine_type]
        if _should_skip_engine(engine_class):
            pytest.skip(
                f"Skipping {engine_class} because it is not supported on this platform"
            )
        mock_ctx = _mock_engine(engine_class)

        with mock_ctx:
            # Build with appropriate params based on engine type
            if issubclass(engine_class, RemoteInferenceEngine):
                engine = build_inference_engine(
                    engine_type=engine_type,
                    model_params=model_params,
                    remote_params=remote_params,
                )
            else:
                engine = build_inference_engine(
                    engine_type=engine_type,
                    model_params=model_params,
                )

            # Verify the engine is of the correct type
            assert isinstance(engine, engine_class)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Optional

from oumi.core.configs.job_config import JobConfig
from oumi.core.launcher.base_cluster import BaseCluster, JobStatus


class BaseCloud(ABC):
    """Base class for resource pool capable of creating clusters."""

    @abstractmethod
    def up_cluster(self, job: JobConfig, name: Optional[str], **kwargs) -> JobStatus:
        """Creates a cluster and starts the provided Job."""
        raise NotImplementedError

    @abstractmethod
    def get_cluster(self, name: str) -> Optional[BaseCluster]:
        """Gets the cluster with the specified name, or None if not found."""
        raise NotImplementedError

    @abstractmethod
    def list_clusters(self) -> list[BaseCluster]:
        """Lists the active clusters on this cloud."""
        raise NotImplementedError



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for AWQ quantization."""

from unittest.mock import MagicMock, patch

import pytest  # type: ignore

from oumi.core.configs import ModelParams, QuantizationConfig
from oumi.quantize.awq_quantizer import AwqQuantization


class TestAwqQuantization:
    """Test cases for AWQ quantization functionality."""

    def setup_method(self):
        """Set up test fixtures."""
        self.awq_quantizer = AwqQuantization()
        self.valid_config = QuantizationConfig(
            model=ModelParams(model_name="TinyLlama/TinyLlama-1.1B-Chat-v1.0"),
            method="awq_q4_0",
            output_path="test_model",
        )

    def test_supported_methods(self):
        """Test AWQ supported methods."""
        expected_methods = ["awq_q4_0", "awq_q4_1", "awq_q8_0", "awq_f16"]
        assert self.awq_quantizer.supported_methods == expected_methods

    def test_supported_formats(self):
        """Test AWQ supported output formats."""
        expected_formats = ["safetensors"]
        assert self.awq_quantizer.supported_formats == expected_formats

    def test_supports_method_valid(self):
        """Test supports_method for valid AWQ methods."""
        assert self.awq_quantizer.supports_method("awq_q4_0") is True
        assert self.awq_quantizer.supports_method("awq_q4_1") is True
        assert self.awq_quantizer.supports_method("awq_q8_0") is True
        assert self.awq_quantizer.supports_method("awq_f16") is True

    def test_supports_method_invalid(self):
        """Test supports_method for invalid methods."""
        assert self.awq_quantizer.supports_method("bnb_4bit") is False
        assert self.awq_quantizer.supports_method("q4_0") is False
        assert self.awq_quantizer.supports_method("invalid") is False

    def test_validate_config_valid(self):
        """Test validate_config with valid AWQ configuration."""
        # Should not raise any exception
        self.awq_quantizer.validate_config(self.valid_config)

    def test_validate_config_invalid_method(self):
        """Test validate_config with non-AWQ method."""
        invalid_config = QuantizationConfig(
            model=ModelParams(model_name="test/model"),
            method="bnb_4bit",
            output_path="test",
        )

        with pytest.raises(ValueError, match="not supported by"):
            self.awq_quantizer.validate_config(invalid_config)

    def test_validate_config_invalid_format(self):
        """Test validate_config with invalid format."""

        with pytest.raises(ValueError, match="Unsupported output format"):
            QuantizationConfig(
                model=ModelParams(model_name="test/model"),
                method="awq_q4_0",
                output_path="test",
                output_format="unknown",
            )

    def test_str_representation(self):
        """Test string representation of AWQ quantizer."""
        assert self.awq_quantizer.__class__.__name__ == "AwqQuantization"

    def test_raise_if_requirements_not_met_missing_awq(self):
        """Test requirements check when AWQ is not installed."""
        # Set _awq to None to simulate missing AWQ
        self.awq_quantizer._awq = None

        with pytest.raises(
            RuntimeError, match="AWQ quantization requires autoawq library"
        ):
            self.awq_quantizer.raise_if_requirements_not_met()

    def test_raise_if_requirements_not_met_no_gpu(self):
        """Test requirements check when no GPU is available."""
        # Set _awq to a mock to simulate AWQ is installed
        self.awq_quantizer._awq = MagicMock()

        with patch("torch.cuda.is_available") as mock_cuda:
            mock_cuda.return_value = False

            with pytest.raises(RuntimeError, match="AWQ quantization requires a GPU"):
                self.awq_quantizer.raise_if_requirements_not_met()


class TestAwqQuantizationSimple:
    """Additional simplified test cases for AWQ quantization."""

    def setup_method(self):
        """Set up test fixtures."""
        self.awq_quantizer = AwqQuantization()

    def test_supports_method(self):
        """Test supports_method for AWQ methods."""
        assert self.awq_quantizer.supports_method("awq_q4_0") is True
        assert self.awq_quantizer.supports_method("awq_q8_0") is True
        assert self.awq_quantizer.supports_method("bnb_4bit") is False

    def test_validate_config_valid_simple(self):
        """Test validate_config with valid configuration."""
        config = QuantizationConfig(
            model=ModelParams(model_name="test/model"),
            method="awq_q4_0",
            output_path="test",
        )
        # Should not raise
        self.awq_quantizer.validate_config(config)

    def test_raise_if_requirements_not_met_no_awq_simple(self):
        """Test requirements check when AWQ is not available."""
        self.awq_quantizer._awq = None
        with pytest.raises(RuntimeError, match="requires autoawq library"):
            self.awq_quantizer.raise_if_requirements_not_met()

    @patch("torch.cuda.is_available")
    def test_raise_if_requirements_not_met_no_gpu_simple(self, mock_cuda):
        """Test requirements check when no GPU is available."""
        mock_cuda.return_value = False
        self.awq_quantizer._awq = MagicMock()  # AWQ is available
        with pytest.raises(RuntimeError, match="requires a GPU"):
            self.awq_quantizer.raise_if_requirements_not_met()
